#include "SIMD.h"

__int64 mmx_cacc(void *A, void *B, int cnt)
{

	int cnt1;
	int cnt2;
	__int64 result;

	cnt1 = cnt / 6;
	cnt2 = (cnt - (6*cnt1));

	__asm 
	{

		//Set up for loop
		mov edi, A;		// Address of A  (IF data)
		mov esi, B;		// Address of B  (complex sine)
		mov ecx, cnt1;	// Counter
		pxor mm0,mm0;	// Clear the running sum
		jecxz ZERO;

		L1:

			movd mm1, [edi]; //load IF data
			movq mm2, [esi]; //load complex sine
	
			movd mm3, [edi+4];
			movq mm4, [esi+8];

			movd mm5, [edi+8];
			movq mm6, [esi+16];

			punpckldq mm1, mm1; //copy bottom 32 bits of IF data into high 32 bits
			punpckldq mm3, mm3; //copy bottom 32 bits of IF data into high 32 bits
			punpckldq mm5, mm5; //copy bottom 32 bits of IF data into high 32 bits
			pmaddwd mm1, mm2;	//perform mmx complex multiply
			pmaddwd mm3, mm4;	//perform mmx complex multiply
			pmaddwd mm5, mm6;	//perform mmx complex multiply
			paddd mm0, mm1;		//add into accumulator
			paddd mm0, mm3;     //add into accumulator
			paddd mm0, mm5;     //add into accumulator

			movd mm1, [edi+12]; //load IF data
			movq mm2, [esi+24]; //load complex sine

			movd mm3, [edi+16];
			movq mm4, [esi+32];

			movd mm5, [edi+20];
			movq mm6, [esi+40];

			punpckldq mm1, mm1; //copy bottom 32 bits of IF data into high 32 bits
			punpckldq mm3, mm3; //copy bottom 32 bits of IF data into high 32 bits
			punpckldq mm5, mm5; //copy bottom 32 bits of IF data into high 32 bits
			pmaddwd mm1, mm2;	//perform mmx complex multiply
			pmaddwd mm3, mm4;	//perform mmx complex multiply
			pmaddwd mm5, mm6;	//perform mmx complex multiply
			paddd mm0, mm1;		//add into accumulator
			paddd mm0, mm3;		//add into accumulator
			paddd mm0, mm5;		//add into accumulator

			add edi, 24;  //move in complex sine by 24 bytes
			add esi, 48;  //move in IF array by 48 bytes

		loop L1;							// Loop if not done

ZERO:
		mov ecx, cnt2;
		jecxz ZERO1;

		L2:

			movd mm1, [edi]; //load IF data
			movq mm2, [esi]; //load complex sine

			punpckldq mm1, mm1; //copy bottom 32 bits of IF data into high 32 bits
			pmaddwd mm1, mm2;	//perform mmx complex multiply
			paddd mm0, mm1;		//add into accumulator

			add edi, 4;  //move in complex sine by 4 bytes
			add esi, 8;  //move in IF array by 8 bytes

		loop L2;	

ZERO1:

		movq result, mm0;

		EMMS;							// done with MMX

	
	}

	return(result);

}


int mmx_dot(void *A, void *B, int cnt)
{

	int temp;
	int temp2 = 0;
	int cnt1;
	int cnt2;
	int cnt3;

	cnt1 = cnt / 28;
	cnt2 = (cnt - (28*cnt1)) / 4;
	cnt3 = (cnt - (28*cnt1) - (4*cnt2));

	__asm 
	{
	
		//Set up for loop
		mov edi, A;		// Address of A
		mov esi, B;		// Address of B
		mov ecx, cnt1;	// Counter
		pxor mm0,mm0;	// Clear the running sum (accumulator)
		jecxz ZERO;


		L1:

			movq mm1, [esi];		//Load from A
			movq mm2, [esi+8];		//Load from A
			movq mm3, [esi+16];		//Load from A
			movq mm4, [esi+24];		//Load from A
			movq mm5, [esi+32];		//Load from A
			movq mm6, [esi+40];		//Load from A
			movq mm7, [esi+48];		//Load from A	6  (13 cycles)

			pmaddwd mm1, [edi];		//Load from B & multiply B * (A*C) (16 cycles, starts on clock 8)
			pmaddwd mm2, [edi+8];	//Load from B & multiply B * (A*C)
			pmaddwd mm3, [edi+16];	//Load from B & multiply B * (A*C)
			pmaddwd mm4, [edi+24];	//Load from B & multiply B * (A*C)
			pmaddwd mm5, [edi+32];	//Load from B & multiply B * (A*C)
			pmaddwd mm6, [edi+40];	//Load from B & multiply B * (A*C)
			pmaddwd mm7, [edi+48];	//Load from B & multiply B * (A*C)	8 (Finished on clock 24)

			paddd mm1, mm2;	//(9 cycles, starts on clock 15)
			paddd mm3, mm4;	
			paddd mm5, mm6;
			paddd mm1, mm3;
			paddd mm0, mm1;
			paddd mm5, mm7;
			paddd mm0, mm5;	//(Finished on clock 24)

			add esi, 56;	// 1
			add edi, 56;	// 1

			loop L1;							// Loop if not done /8 34 cycles

ZERO:
			mov ecx, cnt2;				// Finish off accumulation with second loop
			jecxz ZERO1;

		L2:

			movq mm1, [esi];
			pmaddwd mm1, [edi];
			paddd mm0, mm1;

			add esi, 8;
			add edi, 8;

		loop L2;

			
ZERO1:	

		movd ebx, mm0;					// right-hand word to ebx
		punpckhdq mm0,mm0;				// left-hand word to right side of mm0
		movd eax, mm0;					// left-hand word into eax
		add eax, ebx;					// running sum now in eax
		mov temp, eax;					// move into temp

		mov ecx, cnt3;
		jecxz ZERO2;


		L3:								// Really finish off loop with non SIMD instructions

			mov bx, [edi];				//Move 16 bits into bx
			movsx ebx, bx;				//Sign extend to 32 bits
			mov ax, [esi];				//Move 16 bits into ax
			movsx eax, ax;				//Sign extend to 32 bits
			imul ebx, eax;				//Multiply
			add temp2, ebx;				//Add into accumulator
			add esi, 2;
			add edi, 2;

		loop L3;


ZERO2:

		EMMS;							// done with MMX

	}


	return(temp+temp2);

}


void mmx_add(void *A, void *B, int cnt)
{

	int cnt1;
	int cnt2;
	int cnt3;

	cnt1 = cnt / 32;
	cnt2 = (cnt - (32*cnt1)) / 4;
	cnt3 = (cnt - (32*cnt1) - (4*cnt2));

	__asm 
	{

		//Set up for loop
		mov edi, A;	// Address of A
		mov esi, B;	// Address of B
		mov ecx, cnt1;	// Counter
		jecxz ZERO;

		L1:

			movq mm0, [edi];		//Load from A
			movq mm1, [edi+8];		//Load from A
			movq mm2, [edi+16];		//Load from A
			movq mm3, [edi+24];		//Load from A
			movq mm4, [edi+32];		//Load from A
			movq mm5, [edi+40];		//Load from A
			movq mm6, [edi+48];		//Load from A
			movq mm7, [edi+56];		//Load from A

			paddw mm0, [esi];		//Load from B & add B + A
			paddw mm1, [esi+8];		//Load from B & add B + A
			paddw mm2, [esi+16];	//Load from B & add B + A
			paddw mm3, [esi+24];	//Load from B & add B + A
			paddw mm4, [esi+32];	//Load from B & add B + A
			paddw mm5, [esi+40];	//Load from B & add B + A
			paddw mm6, [esi+48];	//Load from B & add B + A
			paddw mm7, [esi+56];	//Load from B & add B + A

			movq [edi],    mm0;
			movq [edi+8],  mm1;
			movq [edi+16], mm2;
			movq [edi+24], mm3;
			movq [edi+32], mm4;
			movq [edi+40], mm5;
			movq [edi+48], mm6;
			movq [edi+56], mm7;

			add edi, 64;
			add esi, 64;
			
		loop L1;							// Loop if not done

ZERO:
		mov ecx, cnt2;
		jecxz ZERO1;

		L2:

			movq mm1, [edi];		//Load from A
			paddw mm1, [esi];		//Load from B & multiply B * (A*C)
			movq [edi], mm1;
			add edi, 8;
			add esi, 8;

		loop L2;	

ZERO1:

		mov ecx, cnt3;
		jecxz ZERO2;

		mov eax, 0;


		L3:								//Really finish off loop with non SIMD instructions

			mov eax, [edi];
            add eax, [esi];
			mov [edi], ax;
			add esi, 2;
			add edi, 2;

		loop L3;

ZERO2:

		EMMS;

	}

}

void mmx_add(void *A, void *B, void *C, int cnt)
{

	int cnt1;
	int cnt2;
	int cnt3;

	cnt1 = cnt / 32;
	cnt2 = (cnt - (32*cnt1)) / 4;
	cnt3 = (cnt - (32*cnt1) - (4*cnt2));

	__asm 
	{

		//Set up for loop
		mov edi, A;	//Address of A
		mov esi, B;	//Address of B
		mov ebx, C; //Address of C
		mov ecx, cnt1;	//Counter
		jecxz ZERO;

		L1:

			movq mm0, [edi];		//Load from A
			movq mm1, [edi+8];		//Load from A
			movq mm2, [edi+16];		//Load from A
			movq mm3, [edi+24];		//Load from A
			movq mm4, [edi+32];		//Load from A
			movq mm5, [edi+40];		//Load from A
			movq mm6, [edi+48];		//Load from A
			movq mm7, [edi+56];		//Load from A

			paddw mm0, [esi];		//Load from B & add B + A
			paddw mm1, [esi+8];		//Load from B & add B + A
			paddw mm2, [esi+16];	//Load from B & add B + A
			paddw mm3, [esi+24];	//Load from B & add B + A
			paddw mm4, [esi+32];	//Load from B & add B + A
			paddw mm5, [esi+40];	//Load from B & add B + A
			paddw mm6, [esi+48];	//Load from B & add B + A
			paddw mm7, [esi+56];	//Load from B & add B + A

			movq [ebx],    mm0;
			movq [ebx+8],  mm1;
			movq [ebx+16], mm2;
			movq [ebx+24], mm3;
			movq [ebx+32], mm4;
			movq [ebx+40], mm5;
			movq [ebx+48], mm6;
			movq [ebx+56], mm7;

			add edi, 64;
			add esi, 64;
			add ebx, 64;
			
		loop L1;							// Loop if not done

ZERO:
		mov ecx, cnt2;
		jecxz ZERO1;

		L2:

			movq mm1, [edi];		//Load from A
			paddw mm1, [esi];		//Load from B & multiply B * (A*C)
			movq [ebx], mm1;
			add edi, 8;
			add esi, 8;
			add ebx, 8;

		loop L2;	

ZERO1:

		mov ecx, cnt3;
		jecxz ZERO2;

		mov eax, 0;


		L3:								//Really finish off loop with non SIMD instructions

			mov eax, [edi];
            add eax, [esi];
			mov [ebx], ax;
			add esi, 2;
			add edi, 2;
			add ebx, 2;

		loop L3;

ZERO2:

		EMMS;

	}

}


void mmx_mul(void *A, void *B, int cnt)
{

	int cnt1;
	int cnt2;
	int cnt3;

	cnt1 = cnt / 32;
	cnt2 = (cnt - (32*cnt1)) / 4;
	cnt3 = (cnt - (32*cnt1) - (4*cnt2));

	__asm 
	{
		
		//; Set up for loop
		mov edi, A;	// Address of A
		mov esi, B;	// Address of B
		mov ecx, cnt1;	// Counter
		jecxz ZERO;

		L1:

			movq mm0, [edi];		//Load from A
			movq mm1, [edi+8];		//Load from A
			movq mm2, [edi+16];		//Load from A
			movq mm3, [edi+24];		//Load from A
			movq mm4, [edi+32];		//Load from A
			movq mm5, [edi+40];		//Load from A
			movq mm6, [edi+48];		//Load from A
			movq mm7, [edi+56];		//Load from A

			pmullw mm0, [esi];		//Load from B & multiply B*A
			pmullw mm1, [esi+8];	//Load from B & multiply B*A
			pmullw mm2, [esi+16];	//Load from B & multiply B*A
			pmullw mm3, [esi+24];	//Load from B & multiply B*A
			pmullw mm4, [esi+32];	//Load from B & multiply B*A
			pmullw mm5, [esi+40];	//Load from B & multiply B*A
			pmullw mm6, [esi+48];	//Load from B & multiply B*A
			pmullw mm7, [esi+56];	//Load from B & multiply B*A

			movq [edi],    mm0;
			movq [edi+8],  mm1;
			movq [edi+16], mm2;
			movq [edi+24], mm3;
			movq [edi+32], mm4;
			movq [edi+40], mm5;
			movq [edi+48], mm6;
			movq [edi+56], mm7;
	
			add edi, 64;
			add esi, 64;
			
		loop L1;							// Loop if not done

ZERO:

		mov ecx, cnt2;
		jecxz ZERO1;

		L2:

			movq mm1, [edi];		//Load from A
			pmullw mm1, [esi];		//Load from B & multiply B * (A*C)
			movq [edi], mm1;
			add edi, 8;
			add esi, 8;

		loop L2;	

ZERO1:

		mov ecx, cnt3;
		jecxz ZERO2;

		mov eax, 0;


		L3:								//Really finish off loop with non SIMD instructions

			mov eax, [edi];
            imul eax, [esi];
			mov [edi], ax;
			add esi, 2;
			add edi, 2;

		loop L3;

ZERO2:

		EMMS;

	}

}


void mmx_mul(void *A, void *B, void *C, int cnt)
{

	int cnt1;
	int cnt2;
	int cnt3;

	cnt1 = cnt / 32;
	cnt2 = (cnt - (32*cnt1)) / 4;
	cnt3 = (cnt - (32*cnt1) - (4*cnt2));


	__asm 
	{
		
		//; Set up for loop
		mov edi, A;	// Address of A	source1
		mov esi, B;	// Address of B	source2
		mov ebx, C;	// Address of C	dest
		mov ecx, cnt1;	// Counter
		jecxz ZERO;

		L1:

			movq mm0, [edi];		//Load from A
			movq mm1, [edi+8];		//Load from A
			movq mm2, [edi+16];		//Load from A
			movq mm3, [edi+24];		//Load from A
			movq mm4, [edi+32];		//Load from A
			movq mm5, [edi+40];		//Load from A
			movq mm6, [edi+48];		//Load from A
			movq mm7, [edi+56];		//Load from A

			pmullw mm0, [esi];		//Load from B & multiply B * (A*C)
			pmullw mm1, [esi+8];	//Load from B & multiply B * (A*C)
			pmullw mm2, [esi+16];	//Load from B & multiply B * (A*C)
			pmullw mm3, [esi+24];	//Load from B & multiply B * (A*C)
			pmullw mm4, [esi+32];	//Load from B & multiply B * (A*C)
			pmullw mm5, [esi+40];	//Load from B & multiply B * (A*C)
			pmullw mm6, [esi+48];	//Load from B & multiply B * (A*C)
			pmullw mm7, [esi+56];	//Load from B & multiply B * (A*C)

			movq [ebx],	   mm0;		//Store C = A*B
			movq [ebx+8],  mm1;		//Store C = A*B
			movq [ebx+16], mm2;		//Store C = A*B
			movq [ebx+24], mm3;		//Store C = A*B
			movq [ebx+32], mm4;		//Store C = A*B
			movq [ebx+40], mm5;		//Store C = A*B
			movq [ebx+48], mm6;		//Store C = A*B
			movq [ebx+56], mm7;		//Store C = A*B

			add edi, 64;
			add esi, 64;
			add ebx, 64;
			
		loop L1;							// Loop if not done

ZERO:

		mov ecx, cnt2;
		jecxz ZERO1;

		L2:

			movq mm1, [edi];		//Load from A
			pmullw mm1, [esi];		//Load from B & multiply B * (A*C)
			movq [ebx], mm1;
			add edi, 8;
			add esi, 8;
			add ebx, 8;

		loop L2;	

ZERO1:

		mov ecx, cnt3;
		jecxz ZERO2;

		mov eax, 0;


		L3:								//Really finish off loop with non SIMD instructions

			mov eax, [edi];
            imul eax, [esi];
			mov [ebx], ax;
			add esi, 2;
			add edi, 2;
			add ebx, 2;

		loop L3;

ZERO2:

		EMMS;

	}


}



void mmx_sub(void *A, void *B, int cnt)
{

	int cnt1;
	int cnt2;
	int cnt3;

	cnt1 = cnt / 32;
	cnt2 = (cnt - (32*cnt1)) / 4;
	cnt3 = (cnt - (32*cnt1) - (4*cnt2));

	__asm 
	{
		
		//; Set up for loop
		mov edi, A;	// Address of A
		mov esi, B;	// Address of B
		mov ecx, cnt1;	// Counter
		jecxz ZERO;

		L1:

			movq mm0, [edi];		//Load from A
			movq mm1, [edi+8];		//Load from A
			movq mm2, [edi+16];		//Load from A
			movq mm3, [edi+24];		//Load from A
			movq mm4, [edi+32];		//Load from A
			movq mm5, [edi+40];		//Load from A
			movq mm6, [edi+48];		//Load from A
			movq mm7, [edi+56];		//Load from A

			psubw mm0, [esi];		//Load from B & multiply B*A
			psubw mm1, [esi+8];	//Load from B & multiply B*A
			psubw mm2, [esi+16];	//Load from B & multiply B*A
			psubw mm3, [esi+24];	//Load from B & multiply B*A
			psubw mm4, [esi+32];	//Load from B & multiply B*A
			psubw mm5, [esi+40];	//Load from B & multiply B*A
			psubw mm6, [esi+48];	//Load from B & multiply B*A
			psubw mm7, [esi+56];	//Load from B & multiply B*A

			movq [edi], mm0;
			movq [edi+8], mm1;
			movq [edi+16], mm2;
			movq [edi+24], mm3;
			movq [edi+32], mm4;
			movq [edi+40], mm5;
			movq [edi+48], mm6;
			movq [edi+56], mm7;
	
			add edi, 64;
			add esi, 64;
			
		loop L1;							// Loop if not done

ZERO:

		mov ecx, cnt2;
		jecxz ZERO1;

		L2:

			movq mm1, [edi];		//Load from A
			psubw mm1, [esi];		//Load from B & multiply B * (A*C)
			movq [edi], mm1;
			add edi, 8;
			add esi, 8;

		loop L2;	

ZERO1:

		mov ecx, cnt3;
		jecxz ZERO2;

		mov eax, 0;


		L3:								//Really finish off loop with non SIMD instructions

			mov eax, [edi];
            sub eax, [esi];
			mov [edi], ax;
			add esi, 2;
			add edi, 2;

		loop L3;

ZERO2:

		EMMS;

	}

}


void mmx_sub(void *A, void *B, void *C, int cnt)
{

	int cnt1;
	int cnt2;
	int cnt3;

	cnt1 = cnt / 32;
	cnt2 = (cnt - (32*cnt1)) / 4;
	cnt3 = (cnt - (32*cnt1) - (4*cnt2));

	__asm 
	{
		
		//; Set up for loop
		mov edi, A;	// Address of A	source1
		mov esi, B;	// Address of B	source2
		mov ebx, C;	// Address of C	dest
		mov ecx, cnt1;	// Counter
		jecxz ZERO;

		L1:

			movq mm0, [edi];		//Load from A
			movq mm1, [edi+8];		//Load from A
			movq mm2, [edi+16];		//Load from A
			movq mm3, [edi+24];		//Load from A
			movq mm4, [edi+32];		//Load from A
			movq mm5, [edi+40];		//Load from A
			movq mm6, [edi+48];		//Load from A
			movq mm7, [edi+56];		//Load from A

			psubw mm0, [esi];		//Load from B & multiply B * (A*C)
			psubw mm1, [esi+8];	//Load from B & multiply B * (A*C)
			psubw mm2, [esi+16];	//Load from B & multiply B * (A*C)
			psubw mm3, [esi+24];	//Load from B & multiply B * (A*C)
			psubw mm4, [esi+32];	//Load from B & multiply B * (A*C)
			psubw mm5, [esi+40];	//Load from B & multiply B * (A*C)
			psubw mm6, [esi+48];	//Load from B & multiply B * (A*C)
			psubw mm7, [esi+56];	//Load from B & multiply B * (A*C)

			movq [ebx],	   mm0;		//Store C = A*B
			movq [ebx+8],  mm1;		//Store C = A*B
			movq [ebx+16], mm2;		//Store C = A*B
			movq [ebx+24], mm3;		//Store C = A*B
			movq [ebx+32], mm4;		//Store C = A*B
			movq [ebx+40], mm5;		//Store C = A*B
			movq [ebx+48], mm6;		//Store C = A*B
			movq [ebx+56], mm7;		//Store C = A*B

			add edi, 64;
			add esi, 64;
			add ebx, 64;
			
		loop L1;							// Loop if not done

ZERO:

		mov ecx, cnt2;
		jecxz ZERO1;

		L2:

			movq mm1, [edi];		//Load from A
			psubw mm1, [esi];		//Load from B & multiply B * (A*C)
			movq [ebx], mm1;
			add edi, 8;
			add esi, 8;
			add ebx, 8;

		loop L2;	

ZERO1:

		mov ecx, cnt3;
		jecxz ZERO2;

		mov eax, 0;


		L3:								//Really finish off loop with non SIMD instructions

			mov eax, [edi];
            sub eax, [esi];
			mov [ebx], ax;
			add esi, 2;
			add edi, 2;
			add ebx, 2;

		loop L3;

ZERO2:

		EMMS;

	}


}

void mmx_conj(void *A, int cnt)
{

	int cnt1, cnt2;
	

	cnt1 = cnt/14;
	cnt2 = cnt - 14*cnt1;

	short M[4] = {1,-1,1,-1};

	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov ecx, cnt1;		// Counter
		movq mm7, M;		// Move the multiply thingie
		jecxz ZERO1;

		L1:

			movq		mm0, [edi];
			movq		mm1, [edi+8];
			movq		mm2, [edi+16];
			movq		mm3, [edi+24];
			movq		mm4, [edi+32];
			movq		mm5, [edi+40];
			movq		mm6, [edi+48];

			pmullw		mm0, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm1, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm2, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm3, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm4, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm5, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm6, mm7;	//Multiply to get [Re -Im Re -Im]

			movq		[edi],    mm0;	//Move into A
			movq		[edi+8],  mm1;	//Move into A
			movq		[edi+16], mm2;	//Move into A
			movq		[edi+24], mm3;	//Move into A
			movq		[edi+32], mm4;	//Move into A
			movq		[edi+40], mm5;	//Move into A
			movq		[edi+48], mm6;	//Move into A

			add			edi, 56;		//Move in array
		
		loop L1;							// Loop if not done

ZERO1:

		mov ecx, cnt2;		// Counter
		jecxz ZERO2;

		L2:

			movd	mm0, [edi];
			pmullw	mm0, mm7;
			movd	[edi], mm0;

			add		edi, 4;

		loop L2;

ZERO2:

		EMMS;

	}

}


void mmx_conj(void *A, void *C, int cnt)
{

	int cnt1, cnt2;
	

	cnt1 = cnt/14;
	cnt2 = cnt - 14*cnt1;

	short M[4] = {1,-1,1,-1};

	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov esi, C;			// Address of C output1
		mov ecx, cnt1;		// Counter
		movq mm7, M;		// Move the multiply thingie
		jecxz ZERO1;

		L1:

			movq		mm0, [edi];
			movq		mm1, [edi+8];
			movq		mm2, [edi+16];
			movq		mm3, [edi+24];
			movq		mm4, [edi+32];
			movq		mm5, [edi+40];
			movq		mm6, [edi+48];

			pmullw		mm0, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm1, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm2, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm3, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm4, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm5, mm7;	//Multiply to get [Re -Im Re -Im]
			pmullw		mm6, mm7;	//Multiply to get [Re -Im Re -Im]

			movq		[esi],    mm0;	//Move into A
			movq		[esi+8],  mm1;	//Move into A
			movq		[esi+16], mm2;	//Move into A
			movq		[esi+24], mm3;	//Move into A
			movq		[esi+32], mm4;	//Move into A
			movq		[esi+40], mm5;	//Move into A
			movq		[esi+48], mm6;	//Move into A

			add			edi, 56;		//Move in array
			add			esi, 56;		//Move in array
		
		loop L1;							// Loop if not done

ZERO1:

		mov ecx, cnt2;		// Counter
		jecxz ZERO2;

		L2:

			movd	mm0, [edi];
			pmullw	mm0, mm7;
			movd	[esi], mm0;

			add		edi, 4;
			add		esi, 4;

		loop L2;

ZERO2:

		EMMS;

	}

}




void mmx_cmul(void *A, void *B, int cnt, int shift)
{

	int cnt1, cnt2;

	cnt1 = cnt/3;
	cnt2 = cnt - (cnt1*3);

	short M[4] = {1,-1,1,1};

	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov esi, B;			// Address of B	source2
		mov ecx, cnt1;		// Counter
		movq mm7, M;		// Move the multiply thingie
		movd mm6, shift;	// Move the round thingie
		jecxz ZERO1;

		L1:


			movd		mm0, [edi];
			movd		mm1, [edi+4];
			movd		mm2, [edi+8];

			movd		mm3, [esi];
			movd		mm4, [esi+4];
			movd		mm5, [esi+8];

			punpckldq	mm0, mm0;		//Copy low 32 bits to high 32 bits
			punpckldq	mm1, mm1;		//Copy low 32 bits to high 32 bits
			punpckldq	mm2, mm2;		//Copy low 32 bits to high 32 bits

			pshufw		mm3, mm3, 0x14;	//Shuffle to get [Re Im Im Re]
			pshufw		mm4, mm4, 0x14;	//Shuffle to get [Re Im Im Re]
			pshufw		mm5, mm5, 0x14;	//Shuffle to get [Re Im Im Re]

			pmullw		mm3, mm7;	//Multiply to get [Re -Im Im Re]
			pmullw		mm4, mm7;	//Multiply to get [Re -Im Im Re]
			pmullw		mm5, mm7;	//Multiply to get [Re -Im Im Re]

			pmaddwd		mm0, mm3;	//Complex multiply and add
			pmaddwd		mm1, mm4;	//Complex multiply and add
			pmaddwd		mm2, mm5;	//Complex multiply and add

			psrad		mm0, mm6;	//Shift by X bits
			psrad		mm1, mm6;	//Shift by X bits
			psrad		mm2, mm6;	//Shift by X bits

			packssdw	mm0, mm0;	//Get into low 32 bits
			packssdw	mm1, mm1;	//Get into low 32 bits
			packssdw	mm2, mm2;	//Get into low 32 bits

			movd		[edi],   mm0;	//Move into A
			movd		[edi+4], mm1;	//Move into A
			movd		[edi+8], mm2;	//Move into A

			add			edi, 12;		//Move in array
			add			esi, 12;		//Move in array
		
		loop L1;							// Loop if not done

ZERO1:

		mov ecx, cnt2;
		jecxz ZERO2;

L2:

			movd		mm0, [edi];
			movd		mm1, [esi];

			punpckldq	mm0, mm0;		//Copy low 32 bits to high 32 bits
			pshufw		mm1, mm1, 0x14;	//Shuffle to get [Re Im Im Re]
			pmullw		mm1, mm7;		//Multiply to get [Re -Im Im Re]
			pmaddwd		mm0, mm1;		//Complex multiply and add
			psrad		mm0, mm6;		//Shift by X bits
			packssdw	mm0, mm0;		//Get into low 32 bits

			movd		[edi], mm0;	//Move into A

			add edi, 4;
			add esi, 4;

			loop L2;


ZERO2:
		EMMS;

	}


}



void mmx_cmul(void *A, void *B, void *C, int cnt, int shift)
{

	int cnt1, cnt2;

	cnt1 = cnt/3;
	cnt2 = cnt - (cnt1*3);

	short M[4] = {1,-1,1,1};

	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov esi, B;			// Address of B	source2
		mov ebx, C;			// Address of C output1
		mov ecx, cnt1;		// Counter
		movq mm7, M;		// Move the multiply thingie
		movd mm6, shift;	// Move the round thingie
		jecxz ZERO1;

		L1:


			movd		mm0, [edi];
			movd		mm1, [edi+4];
			movd		mm2, [edi+8];

			movd		mm3, [esi];
			movd		mm4, [esi+4];
			movd		mm5, [esi+8];

			punpckldq	mm0, mm0;		//Copy low 32 bits to high 32 bits
			punpckldq	mm1, mm1;		//Copy low 32 bits to high 32 bits
			punpckldq	mm2, mm2;		//Copy low 32 bits to high 32 bits

			pshufw		mm3, mm3, 0x14;	//Shuffle to get [Re Im Im Re]
			pshufw		mm4, mm4, 0x14;	//Shuffle to get [Re Im Im Re]
			pshufw		mm5, mm5, 0x14;	//Shuffle to get [Re Im Im Re]

			pmullw		mm3, mm7;	//Multiply to get [Re -Im Im Re]
			pmullw		mm4, mm7;	//Multiply to get [Re -Im Im Re]
			pmullw		mm5, mm7;	//Multiply to get [Re -Im Im Re]

			pmaddwd		mm0, mm3;	//Complex multiply and add
			pmaddwd		mm1, mm4;	//Complex multiply and add
			pmaddwd		mm2, mm5;	//Complex multiply and add

			psrad		mm0, mm6;	//Shift by X bits
			psrad		mm1, mm6;	//Shift by X bits
			psrad		mm2, mm6;	//Shift by X bits

			packssdw	mm0, mm0;	//Get into low 32 bits
			packssdw	mm1, mm1;	//Get into low 32 bits
			packssdw	mm2, mm2;	//Get into low 32 bits

			movd		[ebx],   mm0;	//Move into C
			movd		[ebx+4], mm1;	//Move into C
			movd		[ebx+8], mm2;	//Move into C

			add			edi, 12;		//Move in array
			add			esi, 12;		//Move in array
			add			ebx, 12;		//Move in array
		
		loop L1;							// Loop if not done

ZERO1:

		mov ecx, cnt2;
		jecxz ZERO2;

L2:

			movd		mm0, [edi];
			movd		mm1, [esi];

			punpckldq	mm0, mm0;		//Copy low 32 bits to high 32 bits
			pshufw		mm1, mm1, 0x14;	//Shuffle to get [Re Im Im Re]
			pmullw		mm1, mm7;		//Multiply to get [Re -Im Im Re]
			pmaddwd		mm0, mm1;		//Complex multiply and add
			psrad		mm0, mm6;		//Shift by X bits
			packssdw	mm0, mm0;		//Get into low 32 bits

			movd		[ebx], mm0;		//Move into C

			add edi, 4;
			add esi, 4;
			add ebx, 4;

			loop L2;


ZERO2:
		EMMS;

	}


}



void mmx_crot(void *A, void *B, int cnt, int shift)
{

	int cnt1;
	int cnt2;

	cnt1 = cnt/5;
	cnt2 = cnt - (cnt1*5);

	short M[4] = {1,-1,1,1};			//To get mmx register into Re Im -Im Re format

	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov esi, B;			// Address of B	source2
		mov ecx, cnt1;		// Counter
		movq mm7, M;		// Move the multiply thingie
		movd mm6, shift;	// Move the round thingie

		//Set up the rotation register
		movd		mm1, [esi];
		pshufw		mm1, mm1, 0x14;	//Shuffle to get [Re Im Im Re]
		pmullw		mm1, mm7;		//Multiply to get [Re -Im Im Re]
		movq		mm7, mm1;		//Now in mm7

		jecxz ZERO1;

		L1:

			movd		mm0, [edi];
			movd		mm1, [edi+4];
			movd		mm2, [edi+8];
			movd		mm3, [edi+12];
			movd		mm4, [edi+16];

			punpckldq	mm0, mm0;	//Copy low 32 bits to high 32 bits
			punpckldq	mm1, mm1;	//Copy low 32 bits to high 32 bits
			punpckldq	mm2, mm2;	//Copy low 32 bits to high 32 bits
			punpckldq	mm3, mm3;	//Copy low 32 bits to high 32 bits
			punpckldq	mm4, mm4;	//Copy low 32 bits to high 32 bits

			pmaddwd		mm0, mm7;	//Complex multiply and add
			pmaddwd		mm1, mm7;	//Complex multiply and add
			pmaddwd		mm2, mm7;	//Complex multiply and add
			pmaddwd		mm3, mm7;	//Complex multiply and add
			pmaddwd		mm4, mm7;	//Complex multiply and add

			psrad		mm0, mm6;	//Shift by X bits
			psrad		mm1, mm6;	//Shift by X bits
			psrad		mm2, mm6;	//Shift by X bits
			psrad		mm3, mm6;	//Shift by X bits
			psrad		mm4, mm6;	//Shift by X bits

			packssdw	mm0, mm0;	//Get into low 32 bits
			packssdw	mm1, mm1;	//Get into low 32 bits
			packssdw	mm2, mm2;	//Get into low 32 bits
			packssdw	mm3, mm3;	//Get into low 32 bits
			packssdw	mm4, mm4;	//Get into low 32 bits

			movd		[edi],    mm0;	//Move into A
			movd		[edi+4],  mm1;	//Move into A
			movd		[edi+8],  mm2;	//Move into A
			movd		[edi+12], mm3;	//Move into A
			movd		[edi+16], mm4;	//Move into A

			add			edi, 20;		//Move in array
		
		loop L1;							// Loop if not done

ZERO1:

		mov ecx, cnt2;
		jecxz ZERO2;

		L2:

			movd		mm0, [edi];

			punpckldq	mm0, mm0;	//Copy low 32 bits to high 32 bits
			pmaddwd		mm0, mm7;	//Complex multiply and add
			psrad		mm0, mm6;	//Shift by X bits
			packssdw	mm0, mm0;	//Get into low 32 bits
			movd		[edi], mm0;	//Move into A

			add			edi, 4;		//Move in array
		
		loop L2;					// Loop if not done

ZERO2:

		EMMS;

	}


}



void mmx_crot(void *A, void *B, void *C, int cnt, int shift)
{

	int cnt1;
	int cnt2;

	cnt1 = cnt/5;
	cnt2 = cnt - (cnt1*5);

	short M[4] = {1,-1,1,1};


	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov esi, B;			// Address of B	source2
		mov ebx, C;			// Address of C output1
		mov ecx, cnt1;		// Counter
		movq mm7, M;		// Move the multiply thingie
		movd mm6, shift;		// Move the shift thingie

		//Set up the rotation register
		movd		mm1, [esi];
		pshufw		mm1, mm1, 0x14;	//Shufflw to get [Re Im Im Re]
		pmullw		mm1, mm7;		//Multiply to get [Re -Im Im Re]
		movq		mm7, mm1;		//Now in mm7

		jecxz ZERO1;

		L1:

			movd		mm0, [edi];
			movd		mm1, [edi+4];
			movd		mm2, [edi+8];
			movd		mm3, [edi+12];
			movd		mm4, [edi+16];

			punpckldq	mm0, mm0;	//Copy low 32 bits to high 32 bits
			punpckldq	mm1, mm1;	//Copy low 32 bits to high 32 bits
			punpckldq	mm2, mm2;	//Copy low 32 bits to high 32 bits
			punpckldq	mm3, mm3;	//Copy low 32 bits to high 32 bits
			punpckldq	mm4, mm4;	//Copy low 32 bits to high 32 bits

			pmaddwd		mm0, mm7;	//Complex multiply and add
			pmaddwd		mm1, mm7;	//Complex multiply and add
			pmaddwd		mm2, mm7;	//Complex multiply and add
			pmaddwd		mm3, mm7;	//Complex multiply and add
			pmaddwd		mm4, mm7;	//Complex multiply and add

			psrad		mm0, mm6;	//Shift by 15 bits
			psrad		mm1, mm6;	//Shift by 15 bits
			psrad		mm2, mm6;	//Shift by 15 bits
			psrad		mm3, mm6;	//Shift by 15 bits
			psrad		mm4, mm6;	//Shift by 15 bits

			packssdw	mm0, mm0;	//Get into low 32 bits
			packssdw	mm1, mm1;	//Get into low 32 bits
			packssdw	mm2, mm2;	//Get into low 32 bits
			packssdw	mm3, mm3;	//Get into low 32 bits
			packssdw	mm4, mm4;	//Get into low 32 bits

			movd		[ebx],	  mm0;	//Move into C
			movd		[ebx+4],  mm1;	//Move into C
			movd		[ebx+8],  mm2;	//Move into C
			movd		[ebx+12], mm3;	//Move into C
			movd		[ebx+16], mm4;	//Move into C

			add			edi, 20;		//Move in array
			add			ebx, 20;
		
		loop L1;							// Loop if not done


ZERO1:

		mov ecx, cnt2;
		jecxz ZERO2;

		L2:

			movd		mm0, [edi];

			punpckldq	mm0, mm0;	//Copy low 32 bits to high 32 bits
			pmaddwd		mm0, mm7;	//Complex multiply and add
			psrad		mm0, mm6;	//Shift by X bits
			packssdw	mm0, mm0;	//Get into low 32 bits
			movd		[ebx], mm0;	//Move into C

			add			edi, 4;		//Move in array
			add			ebx, 4;
		
		loop L2;					// Loop if not done

ZERO2:

		EMMS;

	}


}

void mmx_qnt(void *A, int cnt)
{

	__int64 hash1 = 0x8000800080008000; //to save the sign bit
	__int64 hash2 = 0x0001000100010001; //add the "1" magnitude
	unsigned short	shash1 = 0x8000;
	unsigned short	shash2 = 0xfffe;
	int cnt1;
	int cnt2;

	cnt1 = cnt / 24;						
	cnt2 = (cnt - (24*cnt1));				

	__asm 
	{
		
		// Set up for loop
		mov edi, A;			// Address of A
		mov ecx, cnt1;		// Counter
		movq mm6, [hash1];	//move in the sign bit hash
		movq mm7, [hash2];	//move in the magnitude hash
		jecxz ZERO;


		L1:

			movq mm0, [edi];
			movq mm1, [edi+8];
			movq mm2, [edi+16];
			movq mm3, [edi+24];
			movq mm4, [edi+32];
			movq mm5, [edi+40];

			pand mm0, mm6;			//Clear everything except the sign bit
			pand mm1, mm6;			//Clear everything except the sign bit
			pand mm2, mm6;			//Clear everything except the sign bit
			pand mm3, mm6;			//Clear everything except the sign bit
			pand mm4, mm6;			//Clear everything except the sign bit
			pand mm5, mm6;			//Clear everything except the sign bit

			pcmpeqw mm0, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm1, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm2, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm3, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm4, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm5, mm6;		//Is the sign bit 1 or 0

			por		mm0, mm7;		//Add the magnitude
			por		mm1, mm7;		//Add the magnitude
			por		mm2, mm7;		//Add the magnitude
			por		mm3, mm7;		//Add the magnitude
			por		mm4, mm7;		//Add the magnitude
			por		mm5, mm7;		//Add the magnitude

			movq	[edi],	  mm0;	//Move back to the array
			movq	[edi+8],  mm1;	//Move back to the array
			movq	[edi+16], mm2;	//Move back to the array
			movq	[edi+24], mm3;	//Move back to the array
			movq	[edi+32], mm4;	//Move back to the array
			movq	[edi+40], mm5;	//Move back to the array

			add edi, 48;

		loop L1;					// Loop if not done

ZERO:
			mov ecx, cnt2;			// Finish off operation with second loop
			jecxz ZERO1;

			mov ax, shash1;
			mov bx, shash2;

		L2:

			mov dx, [edi];
			and dx, ax;
			cmp dx, ax;
			cmove dx, bx;
			add dx, 0x0001;
			mov [edi], dx;

			add edi, 2;

		loop L2;

			
ZERO1:	

		EMMS;							// done with MMX

	
	}




}


void mmx_qnt(void *A, void *C, int cnt)
{

	__int64 hash1 = 0x8000800080008000; //to save the sign bit
	__int64 hash2 = 0x0001000100010001; //add the "1" magnitude
	unsigned short	shash1 = 0x8000;
	unsigned short	shash2 = 0xfffe;
	int cnt1;
	int cnt2;

	cnt1 = cnt / 24;						
	cnt2 = (cnt - (24*cnt1));				

	__asm 
	{
		
		// Set up for loop
		mov edi, A;			// Address of A
		mov esi, C;			// Address of C
		mov ecx, cnt1;		// Counter
		movq mm6, [hash1];	//move in the sign bit hash
		movq mm7, [hash2];	//move in the magnitude hash
		jecxz ZERO;


		L1:

			movq mm0, [edi];
			movq mm1, [edi+8];
			movq mm2, [edi+16];
			movq mm3, [edi+24];
			movq mm4, [edi+32];
			movq mm5, [edi+40];

			pand mm0, mm6;			//Clear everything except the sign bit
			pand mm1, mm6;			//Clear everything except the sign bit
			pand mm2, mm6;			//Clear everything except the sign bit
			pand mm3, mm6;			//Clear everything except the sign bit
			pand mm4, mm6;			//Clear everything except the sign bit
			pand mm5, mm6;			//Clear everything except the sign bit

			pcmpeqw mm0, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm1, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm2, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm3, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm4, mm6;		//Is the sign bit 1 or 0
			pcmpeqw mm5, mm6;		//Is the sign bit 1 or 0

			por		mm0, mm7;		//Add the magnitude
			por		mm1, mm7;		//Add the magnitude
			por		mm2, mm7;		//Add the magnitude
			por		mm3, mm7;		//Add the magnitude
			por		mm4, mm7;		//Add the magnitude
			por		mm5, mm7;		//Add the magnitude

			movq	[esi],	  mm0;	//Move back to the array
			movq	[esi+8],  mm1;	//Move back to the array
			movq	[esi+16], mm2;	//Move back to the array
			movq	[esi+24], mm3;	//Move back to the array
			movq	[esi+32], mm4;	//Move back to the array
			movq	[esi+40], mm5;	//Move back to the array

			add edi, 48;
			add esi, 48;

		loop L1;					// Loop if not done

ZERO:
			mov ecx, cnt2;			// Finish off operation with second loop
			jecxz ZERO1;

			mov ax, shash1;
			mov bx, shash2;

		L2:

			mov dx, [edi];
			and dx, ax;
			cmp dx, ax;
			cmove dx, bx;
			add dx, 0x0001;
			mov [esi], dx;

			add edi, 2;
			add esi, 2;

		loop L2;

			
ZERO1:	

		EMMS;							// done with MMX

	
	}




}




bool CPU_MMX()
{

	bool flag;

	_asm
	{
		mov eax, 0x1;
		cpuid;
		sar edx, 0x17;
		and edx, 0x1;
		mov [flag], dl;
	}

	return(flag);

}

bool CPU_SSE()
{

	bool flag;

	_asm
	{
		mov eax, 0x1;
		cpuid;
		sar edx, 0x19;
		and edx, 0x1;
		mov [flag], dl;
	}

	return(flag);

}

bool CPU_SSE2()
{

	bool flag;

	_asm
	{
		mov eax, 0x1;
		cpuid;
		sar edx, 0x1A;
		and edx, 0x1;
		mov [flag], dl;
	}

	return(flag);

}

bool CPU_SSE3()
{

	bool flag;

	_asm
	{
		mov eax, 0x1;
		cpuid;
		and ecx, 0x1;
		mov [flag], cl;
	}

	return(flag);

}


