#include <stdlib.h>
#include <stdio.h>
#include <tchar.h>
#include <math.h>
#include <time.h>
#include <windows.h>

#include "FFT.h"
#include "SIMD.h"

#define ITER 100

int main(int argc, char *argv[])
{

	short *a, *b, *c, *d;
	int accum, sse_accum;
	__int64 sse_caccum;
	int rtick, lcv, len;
	bool err = false;

	//intialize random number
	srand(time(NULL));


	printf("\nTesting sse_mul\n");
	for(rtick = 0; rtick < ITER; rtick++)
	{
		printf(".");
		len = rand()%4096; //size
		a = (short *)malloc(len*sizeof(short));
		b = (short *)malloc(len*sizeof(short));
		c = (short *)malloc(len*sizeof(short));
		d = (short *)malloc(len*sizeof(short));

		//Initialize numbers
		for(lcv = 0; lcv < len; lcv++)
		{
			a[lcv] = rand()%256 - 128;
			b[lcv] = rand()%256 - 128;
		}

		for(lcv = 0; lcv < len; lcv++)
			d[lcv] = a[lcv]*b[lcv];

		sse_mul(a,b,c,len);
		sse_mul(a,b,len);

		for(lcv = 0; lcv < len; lcv++)
		{
			if(a[lcv] != d[lcv])
				err = true;
		}

		for(lcv = 0; lcv < len; lcv++)
		{
			if(c[lcv] != d[lcv])
				err = true;
		}

		free(a);
		free(b);
		free(c);
		free(d);

	}
	if(err)
		printf("\n sse_mul failed!\n");
	else
		printf("\n sse_mul passed!\n");


	printf("\nTesting sse_add\n"); err = false;
	for(rtick = 0; rtick < ITER; rtick++)
	{
		printf(".");
		len = rand()%4096; //size
		a = (short *)malloc(len*sizeof(short));
		b = (short *)malloc(len*sizeof(short));
		c = (short *)malloc(len*sizeof(short));
		d = (short *)malloc(len*sizeof(short));

		//Initialize numbers
		for(lcv = 0; lcv < len; lcv++)
		{
			a[lcv] = rand()%65536 - 32768;
			b [lcv] = rand()%65536 - 32768;
		}

		for(lcv = 0; lcv < len; lcv++)
			d[lcv] = a[lcv]+b[lcv];

		sse_add(a,b,c,len);
		sse_add(a,b,len);

		for(lcv = 0; lcv < len; lcv++)
		{
			if(a[lcv] != d[lcv])
				err = true;
		}

		for(lcv = 0; lcv < len; lcv++)
		{
			if(c[lcv] != d[lcv])
				err = true;
		}

		free(a);
		free(b);
		free(c);
		free(d);

	}

	if(err)
		printf("\n sse_add failed!\n");
	else
		printf("\n sse_add passed!\n");


	printf("\nTesting sse_sub\n"); err = false;
	for(rtick = 0; rtick < ITER; rtick++)
	{
		printf(".");
		len = rand()%4096; //size
		a = (short *)malloc(len*sizeof(short));
		b = (short *)malloc(len*sizeof(short));
		c = (short *)malloc(len*sizeof(short));
		d = (short *)malloc(len*sizeof(short));

		//Initialize numbers
		for(lcv = 0; lcv < len; lcv++)
		{
			a[lcv] = rand()%65536 - 32768;
			b[lcv] = rand()%65536 - 32768;
		}

		for(lcv = 0; lcv < len; lcv++)
			d[lcv] = a[lcv]-b[lcv];

		sse_sub(a,b,c,len);
		sse_sub(a,b,len);

		for(lcv = 0; lcv < len; lcv++)
		{
			if(a[lcv] != d[lcv])
				err = true;
		}

		for(lcv = 0; lcv < len; lcv++)
		{
			if(c[lcv] != d[lcv])
				err = true;
		}

		free(a);
		free(b);
		free(c);
		free(d);

	}

	if(err)
		printf("\n sse_sub failed!\n");
	else
		printf("\n sse_sub passed!\n");


	printf("\nTesting sse_dot\n"); err = false;
	for(rtick = 0; rtick < ITER; rtick++)
	{
		printf(".");
		len = rand()%4096; //size
		a = (short *)malloc(len*sizeof(short));
		b = (short *)malloc(len*sizeof(short));
		accum = 0;

		//Initialize numbers
		for(lcv = 0; lcv < len; lcv++)
		{
			a[lcv] = rand()%257 - 128;
			b[lcv] = rand()%257 - 128;
		}

		for(lcv = 0; lcv < len; lcv++)
			accum += (long)a[lcv]*(long)b[lcv];

		sse_accum = sse_dot(a,b,len);

		if(accum != sse_accum)
			err = true;

		free(a);
		free(b);

	}

	if(err)
	{
		printf("\n sse_dot failed!\n");
	}
	else
		printf("\n sse_dot passed!\n");


	printf("\nTesting sse_cacc\n"); err = false; 
	int raccum; int iaccum;
	for(rtick = 0; rtick < ITER; rtick++)
	{
		printf(".");
		len = rand()%4096; //size
		a = (short *)malloc(2*len*sizeof(short));
		b = (short *)malloc(4*len*sizeof(short));
		raccum = 0;
		iaccum = 0;

		//Initialize numbers
		for(lcv = 0; lcv < len; lcv++)
		{
			a[2*lcv]   = rand()%2 - 1;
			a[2*lcv+1] = rand()%2 - 1;

			b[4*lcv]   = b[4*lcv+3] = rand()%2 - 1;
			b[4*lcv+2] = rand()%2 - 1;
			b[4*lcv+1] = -b[4*lcv+2];
		}

		for(lcv = 0; lcv < len; lcv++)
		{
			raccum += a[2*lcv] * b[4*lcv]   - a[2*lcv+1] * b[4*lcv+2]; 
			iaccum += a[2*lcv] * b[4*lcv+2] + a[2*lcv+1] * b[4*lcv]; 
		}
		
		sse_caccum = sse_cacc(a,b,len);

		if(raccum^(int)(sse_caccum & 0xffffffff))
			err = true;

		if(iaccum^(int)(sse_caccum >> 32))
			err = true;

		free(a);
		free(b);

	}


	if(err)
	{
		printf("\n sse_cacc failed!\n");
	}
	else
		printf("\n sse_cacc passed!\n");


	printf("\nTesting sse_conj\n"); err = false; 
	for(rtick = 0; rtick < ITER; rtick++)
	{
		printf(".");
		len = rand()%4096; //size
		//len = 1;
		a = (short *)malloc(2*len*sizeof(short));
		b = (short *)malloc(2*len*sizeof(short));
		c = (short *)malloc(2*len*sizeof(short));

		//Initialize numbers
		for(lcv = 0; lcv < len; lcv++)
		{
			c[2*lcv]   = a[2*lcv]     = rand()%32768 - 16384;
			c[2*lcv+1] = a[2*lcv+1]   = rand()%32768 - 16384;
		}

		//Perform conjugate
		for(lcv = 0; lcv < len; lcv++)
			c[2*lcv+1] = -c[2*lcv+1];

		sse_conj(a, b, len);
		sse_conj(a, len);

		for(lcv = 0; lcv < 2*len; lcv++)
		{
			if(b[lcv] != c[lcv])
 				err = true;

			if(a[lcv] != c[lcv])
				err = true;
		}

		free(a);
		free(b);
		free(c);

	}

	if(err)
		printf("\n sse_conj failed\n");
	else
		printf("\n sse_conj passed!\n");


	printf("\nTesting sse_qnt\n"); err = false; 
	for(rtick = 0; rtick < ITER; rtick++)
	{
		printf(".");
		len = rand()%4096; //size
		//len = 1;
		a = (short *)malloc(len*sizeof(short));
		b = (short *)malloc(len*sizeof(short));
		c = (short *)malloc(len*sizeof(short));

		//Initialize numbers
		for(lcv = 0; lcv < len; lcv++)
			c[lcv] = a[lcv] = rand()%32768 - 16384;

		//Perform conjugate
		for(lcv = 0; lcv < len; lcv++)
		{
			if(c[lcv] < 0)
				c[lcv] = -1;
			else
				c[lcv] = 1;
		}


		sse_qnt(a, b, len);
		sse_qnt(a, len);

		for(lcv = 0; lcv < len; lcv++)
		{

			if(a[lcv] - c[lcv])
 				err = true;

			if(b[lcv] - c[lcv])
 				err = true;
			
		}

		free(a);
		free(b);
		free(c);

	}

	if(err)
		printf("\n sse_qnt failed\n");
	else
		printf("\n sse_qnt passed!\n");


	printf("\nTesting sse_cmul\n"); err = false; 
	int shift; int tempi; int tempr;
	for(rtick = 0; rtick < ITER; rtick++)
	{
		printf(".");
		len = rand()%4096; //size
		shift = rand()%32; //shift
		//shift = 0;

		//len = 1;
		a = (short *)malloc(2*len*sizeof(short));
		b = (short *)malloc(2*len*sizeof(short));
		c = (short *)malloc(2*len*sizeof(short));
		d = (short *)malloc(2*len*sizeof(short));

		

		//Initialize numbers
		for(lcv = 0; lcv < len; lcv++)
		{
			a[2*lcv]   = rand()%256 - 128;
			a[2*lcv+1] = rand()%256 - 128;

			b[2*lcv] =   rand()%256 - 128;
			b[2*lcv+1] = rand()%256 - 128;
		}

		//Perform multiply
		for(lcv = 0; lcv < len; lcv++)
		{
			tempr = (long)a[2*lcv]*(long)b[2*lcv] - (long)a[2*lcv+1]*(long)b[2*lcv+1];				
			tempi = (long)a[2*lcv+1]*(long)b[2*lcv] + (long)a[2*lcv]*(long)b[2*lcv+1];

			tempr >>= shift;
			tempi >>= shift;

			c[2*lcv] = (short) tempr;
			c[2*lcv+1] =  (short) tempi;

		}


		sse_cmul(a, b, d, len, shift);
		sse_cmul(a, b, len, shift);

		for(lcv = 0; lcv < 2*len; lcv++)
		{

			if(d[lcv] - c[lcv])
 				err = true;

			if(a[lcv] - c[lcv])
 				err = true;
		}

		free(a);
		free(b);
		free(c);
		free(d);

	}

	if(err)
		printf("\n sse_cmul failed\n");
	else
		printf("\n sse_cmul passed!\n");

	printf("\nTesting sse_crot\n"); err = false; 
	for(rtick = 0; rtick < ITER; rtick++)
	{
		printf(".");
		len = rand()%4096; //size
		shift = rand()%32; //shift

		//len = 1;
		a = (short *)malloc(2*len*sizeof(short));
		b = (short *)malloc(2*sizeof(short));
		c = (short *)malloc(2*len*sizeof(short));
		d = (short *)malloc(2*len*sizeof(short));

		b[0] = rand()%256 - 128;
		b[1] = rand()%256 - 128;

		//Initialize numbers
		for(lcv = 0; lcv < len; lcv++)
		{
			a[2*lcv]   = rand()%256 - 128;
			a[2*lcv+1] = rand()%256 - 128;
		}

		//Perform multiply
		for(lcv = 0; lcv < len; lcv++)
		{
			tempr = (long)a[2*lcv]*(long)b[0] - (long)a[2*lcv+1]*(long)b[1];				
			tempi = (long)a[2*lcv+1]*(long)b[0] + (long)a[2*lcv]*(long)b[1];

			tempr >>= shift;
			tempi >>= shift;

			c[2*lcv] = (short) tempr;
			c[2*lcv+1] =  (short) tempi;

		}


		sse_crot(a, b, d, len, shift);
		sse_crot(a, b, len, shift);

		for(lcv = 0; lcv < 2*len; lcv++)
		{

			if(d[lcv] - c[lcv])
 				err = true;

			if(a[lcv] - c[lcv])
 				err = true;
		}

		free(a);
		free(b);
		free(c);
		free(d);

	}

	if(err)
		printf("\n sse_crot failed\n");
	else
		printf("\n sse_crot passed!\n");

	return 0;
}