#include "SIMD.h"

__int64 sse_cacc(void *A, void *B, int cnt)
{

	int cnt1;
	int cnt2;
	__int64 result;


	if(((int)A%16) || ((int)B%16))
	{

		cnt1 = cnt / 6;
		cnt2 = (cnt - (6*cnt1));

		__asm 
		{
		
			//Set up for loop
			mov edi, A;			// Address of A  
			mov esi, B;			// Address of B  
			mov ecx, cnt1;		// Counter
			pxor xmm0, xmm0;	// Clear the running sum
			pxor mm0, mm0;		// Clear the running sum
			jecxz ZERO1;

			L1:

				movlpd xmm1, [edi];		//load IF data
				movlpd xmm2, [edi+8];	//load IF data
				movlpd xmm3, [edi+16];	//load IF data

				movupd xmm4, [esi];		//load Sine data
				movupd xmm5, [esi+16];	//load Sine data
				movupd xmm6, [esi+32];	//load Sine data

				punpckldq xmm1, xmm1;	//copies bits 0..31 to 32..63 and bits 32..63 to 64..95 and 65..127
				punpckldq xmm2, xmm2;	//copies bits 0..63 to 64..127
				punpckldq xmm3, xmm3;	//copies bits 0..63 to 64..127

				pmaddwd	xmm1, xmm4;		//multiply and add, result in xmm1
				pmaddwd xmm2, xmm5;		//multiply and add, result in xmm2
				pmaddwd	xmm3, xmm6;		//multiply and add, result in xmm3

				paddd xmm0, xmm3;		//Add into accumulator (efficiently)
				paddd xmm1, xmm2;		
				paddd xmm0, xmm1;

				add edi, 24;			//move in complex sine by 24 bytes
				add esi, 48;			//move in IF array by 48 bytes

			loop L1;					// Loop if not done

	ZERO1:
			mov ecx, cnt2;
			jecxz ZERO2;

			L2:

				movq		mm1, [edi];		//load IF data

				punpckldq	mm1, mm1;		//copy bottom 32 bits of IF data into high 32 bits
				pmaddwd		mm1, [esi];		//perform mmx complex multiply
				paddd		mm0, mm1;		//add into accumulator

				add edi, 4;					//move in complex sine by 4 bytes
				add esi, 8;					//move in IF array by 8 bytes

			loop L2;	

	ZERO2:

		
			movdq2q		mm1, xmm0;
			punpckhqdq xmm0, xmm0;		//move bits 64..127 of xmm0 into 0..63 of xmm0
			movdq2q		mm2, xmm0;

			paddd		mm0, mm1;		//add together
			paddd		mm0, mm2;		//add together

			movq result, mm0;

			EMMS;						// done with MMX
		
		}//end __asm
	}
	else
	{

		cnt1 = cnt / 12;
		cnt2 = (cnt - (12*cnt1));

		__asm 
		{
		
			//Set up for loop
			mov edi, A;			// Address of A  
			mov esi, B;			// Address of B  
			mov ecx, cnt1;		// Counter
			pxor xmm0, xmm0;	// Clear the running sum
			pxor mm0, mm0;		// Clear the running sum
			jecxz AZERO1;

			AL1:

				movlpd xmm1, [edi];			//load IF data
				movlpd xmm2, [edi+8];		//load IF data
				movlpd xmm3, [edi+16];		//load IF data
				movlpd xmm4, [edi+24];		//load IF data
				movlpd xmm5, [edi+32];		//load IF data
				movlpd xmm6, [edi+40];		//load IF data

				punpckldq xmm1, xmm1;		//copies bits 0..31 to 32..63 and bits 32..63 to 64..95 and 65..127
				punpckldq xmm2, xmm2;		//copies bits 0..63 to 64..127
				punpckldq xmm3, xmm3;		//copies bits 0..63 to 64..127
				punpckldq xmm4, xmm4;		//copies bits 0..63 to 64..127
				punpckldq xmm5, xmm5;		//copies bits 0..63 to 64..127
				punpckldq xmm6, xmm6;		//copies bits 0..63 to 64..127

				pmaddwd	xmm1, [esi];		//multiply and add, result in xmm1
				pmaddwd xmm2, [esi+16];		//multiply and add, result in xmm2
				pmaddwd	xmm3, [esi+32];		//multiply and add, result in xmm3
				pmaddwd	xmm4, [esi+48];		//multiply and add, result in xmm4
				pmaddwd	xmm5, [esi+64];		//multiply and add, result in xmm5
				pmaddwd	xmm6, [esi+80];		//multiply and add, result in xmm6

				paddd	xmm1, xmm2;			//Add into accumulator (efficiently)
				paddd	xmm3, xmm4;
				paddd	xmm5, xmm6;
				paddd	xmm1, xmm3;
				paddd	xmm0, xmm5;
				paddd	xmm0, xmm1;

				add edi, 48;				//move in complex sine by 56 bytes
				add esi, 96;				//move in IF array by 112 bytes

			loop AL1;						// Loop if not done

	AZERO1:
			mov ecx, cnt2;
			jecxz AZERO2;

			AL2:

				movq		mm1, [edi];		//load IF data

				punpckldq	mm1, mm1;		//copy bottom 32 bits of IF data into high 32 bits
				pmaddwd		mm1, [esi];		//perform mmx complex multiply
				paddd		mm0, mm1;		//add into accumulator

				add edi, 4;					//move in complex sine by 4 bytes
				add esi, 8;					//move in IF array by 8 bytes

			loop AL2;	

	AZERO2:

		
			movdq2q		mm1, xmm0;
			punpckhqdq xmm0, xmm0;		//move bits 64..127 of xmm0 into 0..63 of xmm0
			movdq2q		mm2, xmm0;

			paddd		mm0, mm1;		//add together
			paddd		mm0, mm2;		//add together

			movq result, mm0;

			EMMS;						// done with MMX
	
		}//end __asm
	}//end if

	return(result);

}




int sse_dot(void *A, void *B, int cnt)
{

	int temp;
	int temp2 = 0;
	int cnt1;
	int cnt2;
	int cnt3;



	if(((int)A%16) || ((int)B%16))	//If the memory locations are not 16 byte aligned use slower movupd instruction
	{

		cnt1 = cnt / 24;
		cnt2 = (cnt - (24*cnt1)) / 8;
		cnt3 = (cnt - (24*cnt1) - (8*cnt2));

		__asm 
		{
			//Set up for loop
			mov edi, A;			// Address of A
			mov esi, B;			// Address of B
			mov ecx, cnt1;		// Counter
			pxor xmm0, xmm0;	// Clear the running sum (accumulator)
			jecxz ZERO;

			L1:

				movupd xmm1, [esi];			//Load from A
				movupd xmm2, [esi+16];		//Load from A
				movupd xmm3, [esi+32];		//Load from A

				movupd xmm4, [edi];			//Load from B
				movupd xmm5, [edi+16];		//Load from B
				movupd xmm6, [edi+32];		//Load from B				

				pmaddwd xmm1, xmm4;			//Multiply and accumulate
				pmaddwd xmm2, xmm5;			//Multiply and accumulate
				pmaddwd xmm3, xmm6;			//Multiply and accumulate

				paddd	xmm1, xmm3;			//Add into accumulator (efficiently)
				paddd	xmm0, xmm2;
				paddd	xmm0, xmm1;

				add esi, 48;
				add edi, 48;

			loop L1;						// Loop if not done

	ZERO:
				mov ecx, cnt2;				// Finish off accumulation with second loop
				jecxz ZERO1;

			L2:

				movupd	xmm1, [esi];
				movupd	xmm2, [edi];
				pmaddwd xmm1, xmm2;
				paddd	xmm0, xmm1;

				add esi, 16;
				add edi, 16;
				
			loop L2;

				
	ZERO1:	

			
			movd ebx, xmm0;					//right-hand word to ebx
			psrldq xmm0, 4;					//left-hand word to right side of xmm0
			movd eax, xmm0;					//left-hand word into eax
			add eax, ebx;					//running sum now in eax
			mov temp, eax;					//move into temp

			psrldq xmm0, 4;					//left-hand word to right side of xmm0
			movd ebx, xmm0;					//right-hand word to ebx
			psrldq xmm0, 4;					//left-hand word to right side of xmm0
			movd eax, xmm0;					//left-hand word into eax
			add eax, ebx;					//running sum now in eax
			add temp, eax;					//move into temp

			mov ecx, cnt3;
			jecxz ZERO2;

			L3:								//Really finish off loop with non SIMD instructions

				mov bx, [edi];				//Move 16 bits into bx
				movsx ebx, bx;				//Sign extend to 32 bits
				mov ax, [esi];				//Move 16 bits into ax
				movsx eax, ax;				//Sign extend to 32 bits
				imul ebx, eax;				//Multiply
				add temp2, ebx;				//Add into accumulator
				add esi, 2;
				add edi, 2;

			loop L3;


	ZERO2:

			EMMS;							// done with MMX

		}
	}
	else	//use faster movapd instruction
	{

		cnt1 = cnt / 56;
		cnt2 = (cnt - (56*cnt1)) / 8;
		cnt3 = (cnt - (56*cnt1) - (8*cnt2));

		__asm 
		{
			//Set up for loop
			mov edi, A;		// Address of A
			mov esi, B;		// Address of B
			mov ecx, cnt1;	// Counter
			pxor xmm0, xmm0;	// Clear the running sum (accumulator)
			jecxz AZERO;

			AL1:

				movapd xmm1, [esi];			//Load from A
				movapd xmm2, [esi+16];		//Load from A
				movapd xmm3, [esi+32];		//Load from A
				movapd xmm4, [esi+48];		//Load from A
				movapd xmm5, [esi+64];		//Load from A
				movapd xmm6, [esi+80];		//Load from A
				movapd xmm7, [esi+96];		//Load from A

				pmaddwd xmm1, [edi];
				pmaddwd xmm2, [edi+16];
				pmaddwd xmm3, [edi+32];
				pmaddwd xmm4, [edi+48];
				pmaddwd xmm5, [edi+64];
				pmaddwd xmm6, [edi+80];
				pmaddwd xmm7, [edi+96];

				paddd	xmm0, xmm7;
				paddd	xmm1, xmm2;
				paddd	xmm3, xmm4;
				paddd	xmm5, xmm6;
				paddd	xmm1, xmm3;
				paddd	xmm0, xmm5;
				paddd	xmm0, xmm1;

				add esi, 112;
				add edi, 112;

			loop AL1;						// Loop if not done

	AZERO:
				mov ecx, cnt2;				// Finish off accumulation with second loop
				jecxz AZERO1;

			AL2:

				movapd xmm1, [esi];
				pmaddwd xmm1, [edi];
				paddd xmm0, xmm1;
				add esi, 16;
				add edi, 16;
				
			loop AL2;

				
	AZERO1:	

			
			movd ebx, xmm0;					// right-hand word to ebx
			psrldq xmm0, 4;					// left-hand word to right side of xmm0
			movd eax, xmm0;					// left-hand word into eax
			add eax, ebx;					// running sum now in eax
			mov temp, eax;					// move into temp

			psrldq xmm0, 4;
			movd ebx, xmm0;					// right-hand word to ebx
			psrldq xmm0, 4;					// left-hand word to right side of xmm0
			movd eax, xmm0;					// left-hand word into eax
			add eax, ebx;					// running sum now in eax
			add temp, eax;					// add into temp


		mov ecx, cnt3;
		jecxz AZERO2;

			mov ecx, cnt3;
			jecxz AZERO2;

			AL3:								//Really finish off loop with non SIMD instructions
			
				mov bx, [edi];				//Move 16 bits into bx
				movsx ebx, bx;				//Sign extend to 32 bits
				mov ax, [esi];				//Move 16 bits into ax
				movsx eax, ax;				//Sign extend to 32 bits
				imul ebx, eax;				//Multiply
				add temp2, ebx;				//Add into accumulator
				add esi, 2;
				add edi, 2;

			loop AL3;

AZERO2:

			EMMS;							// done with MMX

		}
	}

	temp += temp2;

	return(temp);

}


void sse_add(void *A, void *B, int cnt)
{

	//MOV EAX,1               ;request CPU feature flags
	//CPUID                   ;0Fh, 0A2h CPUID instruction
	//TEST EDX,4000000h       ;test bit 26 (SSE2)
	//JNZ >L18                ;SSE2 available

	int cnt1;
	int cnt2;
	int cnt3;

	if(((int)A%16) || ((int)B%16)) //unaligned version
	{

		cnt1 = cnt / 32;
		cnt2 = (cnt - (32*cnt1)) / 8;
		cnt3 = (cnt - (32*cnt1) - (8*cnt2));

		_asm 
		{
		
			//Set up for loop
			mov edi, A;	// Address of A
			mov esi, B;	// Address of B
			mov ecx, cnt1;	// Counter
			jecxz ZERO;

			L1:

				movupd xmm0, [edi];			//Load from A
				movupd xmm1, [edi+16];		//Load from A
				movupd xmm2, [edi+32];		//Load from A
				movupd xmm3, [edi+48];		//Load from A

				movupd xmm4, [esi];			//Load from B
				movupd xmm5, [esi+16];		//Load from B
				movupd xmm6, [esi+32];		//Load from B
				movupd xmm7, [esi+48];		//Load from B

				paddw xmm0, xmm4;			//Multiply A*B
				paddw xmm1, xmm5;			//Multiply A*B
				paddw xmm2, xmm6;			//Multiply A*B
				paddw xmm3, xmm7;			//Multiply A*B

				movupd [edi],	  xmm0;		//Move into A
				movupd [edi+16],  xmm1;		//Move into A
				movupd [edi+32],  xmm2;		//Move into A
				movupd [edi+48],  xmm3;		//Move into A

				add edi, 64;
				add esi, 64;
				
			loop L1;							// Loop if not done

	ZERO:
			mov ecx, cnt2;
			jecxz ZERO1;

			L2:

				movupd xmm0, [edi];		//Load from A
				movupd xmm1, [esi];		//Load from B
				paddw xmm0, xmm1;		//Load from B & multiply A*B
				movupd [edi], xmm0;		//Move into A
				add edi, 16;
				add esi, 16;

			loop L2;	

	ZERO1:

			mov ecx, cnt3;
			jecxz ZERO2;

			mov eax, 0;

			L3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				add ax, [esi];
				mov [edi], ax;
				add esi, 2;
				add edi, 2;

			loop L3;

	ZERO2:

			EMMS;

		}//end __asm

	}
	else
	{

		cnt1 = cnt / 56;
		cnt2 = (cnt - (56*cnt1)) / 8;
		cnt3 = (cnt - (56*cnt1) - (8*cnt2));

		_asm 
		{
		
			//Set up for loop
			mov edi, A;	// Address of A
			mov esi, B;	// Address of B
			mov ecx, cnt1;	// Counter
			jecxz AZERO;

			AL1:

				movapd xmm0, [edi];		//Load from A
				movapd xmm1, [edi+16];		//Load from A
				movapd xmm2, [edi+32];		//Load from A
				movapd xmm3, [edi+48];		//Load from A
				movapd xmm4, [edi+64];		//Load from A
				movapd xmm5, [edi+80];		//Load from A
				movapd xmm6, [edi+96];		//Load from A

				paddw xmm0, [esi];			//Load from B & multiply A*B
				paddw xmm1, [esi+16];		//Load from B & multiply A*B
				paddw xmm2, [esi+32];		//Load from B & multiply A*B
				paddw xmm3, [esi+48];		//Load from B & multiply A*B
				paddw xmm4, [esi+64];		//Load from B & multiply A*B
				paddw xmm5, [esi+80];		//Load from B & multiply A*B
				paddw xmm6, [esi+96];		//Load from B & multiply A*B

				movapd [edi],	  xmm0;		//Move into A
				movapd [edi+16],  xmm1;		//Move into A
				movapd [edi+32],  xmm2;		//Move into A
				movapd [edi+48],  xmm3;		//Move into A
				movapd [edi+64],  xmm4;		//Move into A
				movapd [edi+80],  xmm5;		//Move into A
				movapd [edi+96],  xmm6;		//Move into A

				add edi, 112;
				add esi, 112;
				
			loop AL1;							// Loop if not done

	AZERO:
			mov ecx, cnt2;
			jecxz AZERO1;

			AL2:

				movapd xmm0, [edi];		//Load from A
				paddw xmm0, [esi];		//Load from B & multiply A*B
				movapd [edi], xmm0;		//Move into A
				add edi, 16;
				add esi, 16;

			loop AL2;	

	AZERO1:

			mov ecx, cnt3;
			jecxz AZERO2;

			mov eax, 0;

			AL3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				add ax, [esi];
				mov [edi], ax;
				add esi, 2;
				add edi, 2;

			loop AL3;

	AZERO2:

			EMMS;

		}//end __asm

	}//end if
}
















void sse_add(void *A, void *B, void *C, int cnt)
{

	int cnt1;
	int cnt2;
	int cnt3;


	if(((int)A%16) || ((int)B%16) || ((int)C%16))
	{

		cnt1 = cnt / 32;
		cnt2 = (cnt - (32*cnt1)) / 8;
		cnt3 = (cnt - (32*cnt1) - (8*cnt2));

		_asm 
		{
		
			//Set up for loop
			mov edi, A;	// Address of A, input1
			mov esi, B;	// Address of B, input2
			mov ebx, C; // Address of C, output1
			mov ecx, cnt1;	// Counter
			jecxz ZERO;

			L1:


				movupd xmm0, [edi];			//Load from A
				movupd xmm1, [edi+16];		//Load from A
				movupd xmm2, [edi+32];		//Load from A
				movupd xmm3, [edi+48];		//Load from A
				movupd xmm4, [esi];			//Load from B
				movupd xmm5, [esi+16];		//Load from B
				movupd xmm6, [esi+32];		//Load from B
				movupd xmm7, [esi+48];		//Load from B

				paddw xmm0, xmm4;			//Multiply A*B
				paddw xmm1, xmm5;			//Multiply A*B
				paddw xmm2, xmm6;			//Multiply A*B
				paddw xmm3, xmm7;			//Multiply A*B

				movupd [ebx],	  xmm0;		//Move into C
				movupd [ebx+16],  xmm1;		//Move into C
				movupd [ebx+32],  xmm2;		//Move into C
				movupd [ebx+48],  xmm3;		//Move into C

				add edi, 64;
				add esi, 64;
				add ebx, 64;
				
			loop L1;							// Loop if not done

	ZERO:
			mov ecx, cnt2;
			jecxz ZERO1;

			L2:

				movupd xmm0, [edi];		//Load from A
				movupd xmm1, [esi];		//Load from B
				paddw  xmm0, xmm1;		//Multiply A*B
				movupd [ebx], xmm0;		//Move into C
				add edi, 16;
				add esi, 16;
				add ebx, 16;

			loop L2;	

	ZERO1:

			mov ecx, cnt3;
			jecxz ZERO2;

			mov eax, 0;

			L3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				add ax, [esi];
				mov [ebx], ax;
				add esi, 2;
				add edi, 2;
				add ebx, 2;

			loop L3;

	ZERO2:

			EMMS;

		}//end __asm

	}
	else
	{

		cnt1 = cnt / 56;
		cnt2 = (cnt - (56*cnt1)) / 8;
		cnt3 = (cnt - (56*cnt1) - (8*cnt2));

		_asm 
		{
			//Set up for loop
			mov edi, A;	// Address of A, input1
			mov esi, B;	// Address of B, input2
			mov ebx, C; // Address of C, output1
			mov ecx, cnt1;	// Counter
			jecxz AZERO;

			AL1:

				movapd	xmm0, [edi];		//Load from A
				movapd	xmm1, [edi+16];		//Load from A
				movapd	xmm2, [edi+32];		//Load from A
				movapd	xmm3, [edi+48];		//Load from A
				movapd	xmm4, [edi+64];		//Load from A
				movapd	xmm5, [edi+80];		//Load from A
				movapd	xmm6, [edi+96];		//Load from A

				paddw xmm0, [esi];			//Load from B & multiply A*B
				paddw xmm1, [esi+16];		//Load from B & multiply A*B
				paddw xmm2, [esi+32];		//Load from B & multiply A*B
				paddw xmm3, [esi+48];		//Load from B & multiply A*B
				paddw xmm4, [esi+64];		//Load from B & multiply A*B
				paddw xmm5, [esi+80];		//Load from B & multiply A*B
				paddw xmm6, [esi+96];		//Load from B & multiply A*B

				movapd [ebx],	  xmm0;		//Move into A
				movapd [ebx+16],  xmm1;		//Move into A
				movapd [ebx+32],  xmm2;		//Move into A
				movapd [ebx+48],  xmm3;		//Move into A
				movapd [ebx+64],  xmm4;		//Move into A
				movapd [ebx+80],  xmm5;		//Move into A
				movapd [ebx+96],  xmm6;		//Move into A

				add edi, 112;
				add esi, 112;
				add ebx, 112;
				
			loop AL1;							// Loop if not done

	AZERO:
			mov ecx, cnt2;
			jecxz AZERO1;

			AL2:

				movapd  xmm0, [edi];		//Load from A
				paddw  xmm0, [esi];		//Load from B & add A+B
				movapd [ebx], xmm0;		//Move into A
				add edi, 16;
				add esi, 16;
				add ebx, 16;

			loop AL2;	

	AZERO1:

			mov ecx, cnt3;
			jecxz AZERO2;

			mov eax, 0;

			AL3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				add ax, [esi];
				mov [ebx], ax;
				add esi, 2;
				add edi, 2;
				add ebx, 2;

			loop AL3;

	AZERO2:

			EMMS;

		}//end __asm

	}//end if
}
















void sse_mul(void *A, void *B, int cnt)
{

	//MOV EAX,1               ;request CPU feature flags
	//CPUID                   ;0Fh, 0A2h CPUID instruction
	//TEST EDX,4000000h       ;test bit 26 (SSE2)
	//JNZ >L18                ;SSE2 available

	int cnt1;
	int cnt2;
	int cnt3;

	if(((int)A%16) || ((int)B%16)) //unaligned version
	{

		cnt1 = cnt / 32;
		cnt2 = (cnt - (32*cnt1)) / 8;
		cnt3 = (cnt - (32*cnt1) - (8*cnt2));

		_asm 
		{
		
			//Set up for loop
			mov edi, A;	// Address of A
			mov esi, B;	// Address of B
			mov ecx, cnt1;	// Counter
			jecxz ZERO;

			L1:

				movupd xmm0, [edi];			//Load from A
				movupd xmm1, [edi+16];		//Load from A
				movupd xmm2, [edi+32];		//Load from A
				movupd xmm3, [edi+48];		//Load from A

				movupd xmm4, [esi];			//Load from B
				movupd xmm5, [esi+16];		//Load from B
				movupd xmm6, [esi+32];		//Load from B
				movupd xmm7, [esi+48];		//Load from B

				pmullw xmm0, xmm4;			//Multiply A*B
				pmullw xmm1, xmm5;			//Multiply A*B
				pmullw xmm2, xmm6;			//Multiply A*B
				pmullw xmm3, xmm7;			//Multiply A*B

				movupd [edi],	  xmm0;		//Move into A
				movupd [edi+16],  xmm1;		//Move into A
				movupd [edi+32],  xmm2;		//Move into A
				movupd [edi+48],  xmm3;		//Move into A

				add edi, 64;
				add esi, 64;
				
			loop L1;							// Loop if not done

	ZERO:
			mov ecx, cnt2;
			jecxz ZERO1;

			L2:

				movupd xmm0, [edi];		//Load from A
				movupd xmm1, [esi];		//Load from B
				pmullw xmm0, xmm1;		//Load from B & multiply A*B
				movupd [edi], xmm0;		//Move into A
				add edi, 16;
				add esi, 16;

			loop L2;	

	ZERO1:

			mov ecx, cnt3;
			jecxz ZERO2;

			mov eax, 0;

			L3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				imul ax, [esi];
				mov [edi], ax;
				add esi, 2;
				add edi, 2;

			loop L3;

	ZERO2:

			EMMS;

		}//end __asm

	}
	else
	{

		cnt1 = cnt / 56;
		cnt2 = (cnt - (56*cnt1)) / 8;
		cnt3 = (cnt - (56*cnt1) - (8*cnt2));

		_asm 
		{
		
			//Set up for loop
			mov edi, A;	// Address of A
			mov esi, B;	// Address of B
			mov ecx, cnt1;	// Counter
			jecxz AZERO;

			AL1:

				movapd xmm0, [edi];		//Load from A
				movapd xmm1, [edi+16];		//Load from A
				movapd xmm2, [edi+32];		//Load from A
				movapd xmm3, [edi+48];		//Load from A
				movapd xmm4, [edi+64];		//Load from A
				movapd xmm5, [edi+80];		//Load from A
				movapd xmm6, [edi+96];		//Load from A

				pmullw xmm0, [esi];			//Load from B & multiply A*B
				pmullw xmm1, [esi+16];		//Load from B & multiply A*B
				pmullw xmm2, [esi+32];		//Load from B & multiply A*B
				pmullw xmm3, [esi+48];		//Load from B & multiply A*B
				pmullw xmm4, [esi+64];		//Load from B & multiply A*B
				pmullw xmm5, [esi+80];		//Load from B & multiply A*B
				pmullw xmm6, [esi+96];		//Load from B & multiply A*B

				movapd [edi],	  xmm0;		//Move into A
				movapd [edi+16],  xmm1;		//Move into A
				movapd [edi+32],  xmm2;		//Move into A
				movapd [edi+48],  xmm3;		//Move into A
				movapd [edi+64],  xmm4;		//Move into A
				movapd [edi+80],  xmm5;		//Move into A
				movapd [edi+96],  xmm6;		//Move into A

				add edi, 112;
				add esi, 112;
				
			loop AL1;							// Loop if not done

	AZERO:
			mov ecx, cnt2;
			jecxz AZERO1;

			AL2:

				movapd xmm0, [edi];		//Load from A
				pmullw xmm0, [esi];		//Load from B & multiply A*B
				movapd [edi], xmm0;		//Move into A
				add edi, 16;
				add esi, 16;

			loop AL2;	

	AZERO1:

			mov ecx, cnt3;
			jecxz AZERO2;

			mov eax, 0;

			AL3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				imul ax, [esi];
				mov [edi], ax;
				add esi, 2;
				add edi, 2;

			loop AL3;

	AZERO2:

			EMMS;

		}//end __asm

	}//end if
}
















void sse_mul(void *A, void *B, void *C, int cnt)
{

	int cnt1;
	int cnt2;
	int cnt3;


	if(((int)A%16) || ((int)B%16) || ((int)C%16))
	{

		cnt1 = cnt / 32;
		cnt2 = (cnt - (32*cnt1)) / 8;
		cnt3 = (cnt - (32*cnt1) - (8*cnt2));

		_asm 
		{
		
			//Set up for loop
			mov edi, A;	// Address of A, input1
			mov esi, B;	// Address of B, input2
			mov ebx, C; // Address of C, output1
			mov ecx, cnt1;	// Counter
			jecxz ZERO;

			L1:


				movupd xmm0, [edi];			//Load from A
				movupd xmm1, [edi+16];		//Load from A
				movupd xmm2, [edi+32];		//Load from A
				movupd xmm3, [edi+48];		//Load from A
				movupd xmm4, [esi];			//Load from B
				movupd xmm5, [esi+16];		//Load from B
				movupd xmm6, [esi+32];		//Load from B
				movupd xmm7, [esi+48];		//Load from B

				pmullw xmm0, xmm4;			//Multiply A*B
				pmullw xmm1, xmm5;			//Multiply A*B
				pmullw xmm2, xmm6;			//Multiply A*B
				pmullw xmm3, xmm7;			//Multiply A*B

				movupd [ebx],	  xmm0;		//Move into C
				movupd [ebx+16],  xmm1;		//Move into C
				movupd [ebx+32],  xmm2;		//Move into C
				movupd [ebx+48],  xmm3;		//Move into C

				add edi, 64;
				add esi, 64;
				add ebx, 64;
				
			loop L1;							// Loop if not done

	ZERO:
			mov ecx, cnt2;
			jecxz ZERO1;

			L2:

				movupd xmm0, [edi];		//Load from A
				movupd xmm1, [esi];		//Load from B
				pmullw  xmm0, xmm1;		//Multiply A*B
				movupd [ebx], xmm0;		//Move into C
				add edi, 16;
				add esi, 16;
				add ebx, 16;

			loop L2;	

	ZERO1:

			mov ecx, cnt3;
			jecxz ZERO2;

			mov eax, 0;

			L3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				imul ax, [esi];
				mov [ebx], ax;
				add esi, 2;
				add edi, 2;
				add ebx, 2;

			loop L3;

	ZERO2:

			EMMS;

		}//end __asm

	}
	else
	{

		cnt1 = cnt / 56;
		cnt2 = (cnt - (56*cnt1)) / 8;
		cnt3 = (cnt - (56*cnt1) - (8*cnt2));

		_asm 
		{
			//Set up for loop
			mov edi, A;	// Address of A, input1
			mov esi, B;	// Address of B, input2
			mov ebx, C; // Address of C, output1
			mov ecx, cnt1;	// Counter
			jecxz AZERO;

			AL1:

				movapd	xmm0, [edi];		//Load from A
				movapd	xmm1, [edi+16];		//Load from A
				movapd	xmm2, [edi+32];		//Load from A
				movapd	xmm3, [edi+48];		//Load from A
				movapd	xmm4, [edi+64];		//Load from A
				movapd	xmm5, [edi+80];		//Load from A
				movapd	xmm6, [edi+96];		//Load from A

				pmullw xmm0, [esi];			//Load from B & multiply A*B
				pmullw xmm1, [esi+16];		//Load from B & multiply A*B
				pmullw xmm2, [esi+32];		//Load from B & multiply A*B
				pmullw xmm3, [esi+48];		//Load from B & multiply A*B
				pmullw xmm4, [esi+64];		//Load from B & multiply A*B
				pmullw xmm5, [esi+80];		//Load from B & multiply A*B
				pmullw xmm6, [esi+96];		//Load from B & multiply A*B

				movapd [ebx],	  xmm0;		//Move into A
				movapd [ebx+16],  xmm1;		//Move into A
				movapd [ebx+32],  xmm2;		//Move into A
				movapd [ebx+48],  xmm3;		//Move into A
				movapd [ebx+64],  xmm4;		//Move into A
				movapd [ebx+80],  xmm5;		//Move into A
				movapd [ebx+96],  xmm6;		//Move into A

				add edi, 112;
				add esi, 112;
				add ebx, 112;
				
			loop AL1;							// Loop if not done

	AZERO:
			mov ecx, cnt2;
			jecxz AZERO1;

			AL2:

				movapd  xmm0, [edi];		//Load from A
				pmullw  xmm0, [esi];		//Load from B & add A+B
				movapd [ebx], xmm0;		//Move into A
				add edi, 16;
				add esi, 16;
				add ebx, 16;

			loop AL2;	

	AZERO1:

			mov ecx, cnt3;
			jecxz AZERO2;

			mov eax, 0;

			AL3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				imul ax, [esi];
				mov [ebx], ax;
				add esi, 2;
				add edi, 2;
				add ebx, 2;

			loop AL3;

	AZERO2:

			EMMS;

		}//end __asm

	}//end if
}















void sse_sub(void *A, void *B, int cnt)
{

	//MOV EAX,1               ;request CPU feature flags
	//CPUID                   ;0Fh, 0A2h CPUID instruction
	//TEST EDX,4000000h       ;test bit 26 (SSE2)
	//JNZ >L18                ;SSE2 available

	int cnt1;
	int cnt2;
	int cnt3;

	if(((int)A%16) || ((int)B%16)) //unaligned version
	{

		cnt1 = cnt / 32;
		cnt2 = (cnt - (32*cnt1)) / 8;
		cnt3 = (cnt - (32*cnt1) - (8*cnt2));

		_asm 
		{
		
			//Set up for loop
			mov edi, A;	// Address of A
			mov esi, B;	// Address of B
			mov ecx, cnt1;	// Counter
			jecxz ZERO;

			L1:

				movupd xmm0, [edi];			//Load from A
				movupd xmm1, [edi+16];		//Load from A
				movupd xmm2, [edi+32];		//Load from A
				movupd xmm3, [edi+48];		//Load from A

				movupd xmm4, [esi];			//Load from B
				movupd xmm5, [esi+16];		//Load from B
				movupd xmm6, [esi+32];		//Load from B
				movupd xmm7, [esi+48];		//Load from B

				psubw xmm0, xmm4;			//Multiply A*B
				psubw xmm1, xmm5;			//Multiply A*B
				psubw xmm2, xmm6;			//Multiply A*B
				psubw xmm3, xmm7;			//Multiply A*B

				movupd [edi],	  xmm0;		//Move into A
				movupd [edi+16],  xmm1;		//Move into A
				movupd [edi+32],  xmm2;		//Move into A
				movupd [edi+48],  xmm3;		//Move into A

				add edi, 64;
				add esi, 64;
				
			loop L1;							// Loop if not done

	ZERO:
			mov ecx, cnt2;
			jecxz ZERO1;

			L2:

				movupd xmm0, [edi];		//Load from A
				movupd xmm1, [esi];		//Load from B
				psubw xmm0, xmm1;		//Load from B & multiply A*B
				movupd [edi], xmm0;		//Move into A
				add edi, 16;
				add esi, 16;

			loop L2;	

	ZERO1:

			mov ecx, cnt3;
			jecxz ZERO2;

			mov eax, 0;

			L3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				sub ax, [esi];
				mov [edi], ax;
				add esi, 2;
				add edi, 2;

			loop L3;

	ZERO2:

			EMMS;

		}//end __asm

	}
	else
	{

		cnt1 = cnt / 56;
		cnt2 = (cnt - (56*cnt1)) / 8;
		cnt3 = (cnt - (56*cnt1) - (8*cnt2));

		_asm 
		{
		
			//Set up for loop
			mov edi, A;	// Address of A
			mov esi, B;	// Address of B
			mov ecx, cnt1;	// Counter
			jecxz AZERO;

			AL1:

				movapd xmm0, [edi];		//Load from A
				movapd xmm1, [edi+16];		//Load from A
				movapd xmm2, [edi+32];		//Load from A
				movapd xmm3, [edi+48];		//Load from A
				movapd xmm4, [edi+64];		//Load from A
				movapd xmm5, [edi+80];		//Load from A
				movapd xmm6, [edi+96];		//Load from A

				psubw xmm0, [esi];			//Load from B & multiply A*B
				psubw xmm1, [esi+16];		//Load from B & multiply A*B
				psubw xmm2, [esi+32];		//Load from B & multiply A*B
				psubw xmm3, [esi+48];		//Load from B & multiply A*B
				psubw xmm4, [esi+64];		//Load from B & multiply A*B
				psubw xmm5, [esi+80];		//Load from B & multiply A*B
				psubw xmm6, [esi+96];		//Load from B & multiply A*B

				movapd [edi],	  xmm0;		//Move into A
				movapd [edi+16],  xmm1;		//Move into A
				movapd [edi+32],  xmm2;		//Move into A
				movapd [edi+48],  xmm3;		//Move into A
				movapd [edi+64],  xmm4;		//Move into A
				movapd [edi+80],  xmm5;		//Move into A
				movapd [edi+96],  xmm6;		//Move into A

				add edi, 112;
				add esi, 112;
				
			loop AL1;							// Loop if not done

	AZERO:
			mov ecx, cnt2;
			jecxz AZERO1;

			AL2:

				movapd xmm0, [edi];		//Load from A
				psubw xmm0, [esi];		//Load from B & multiply A*B
				movapd [edi], xmm0;		//Move into A
				add edi, 16;
				add esi, 16;

			loop AL2;	

	AZERO1:

			mov ecx, cnt3;
			jecxz AZERO2;

			mov eax, 0;

			AL3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				sub ax, [esi];
				mov [edi], ax;
				add esi, 2;
				add edi, 2;

			loop AL3;

	AZERO2:

			EMMS;

		}//end __asm

	}//end if
}
















void sse_sub(void *A, void *B, void *C, int cnt)
{

	int cnt1;
	int cnt2;
	int cnt3;


	if(((int)A%16) || ((int)B%16) || ((int)C%16))
	{

		cnt1 = cnt / 32;
		cnt2 = (cnt - (32*cnt1)) / 8;
		cnt3 = (cnt - (32*cnt1) - (8*cnt2));

		_asm 
		{
		
			//Set up for loop
			mov edi, A;	// Address of A, input1
			mov esi, B;	// Address of B, input2
			mov ebx, C; // Address of C, output1
			mov ecx, cnt1;	// Counter
			jecxz ZERO;

			L1:


				movupd xmm0, [edi];			//Load from A
				movupd xmm1, [edi+16];		//Load from A
				movupd xmm2, [edi+32];		//Load from A
				movupd xmm3, [edi+48];		//Load from A
				movupd xmm4, [esi];			//Load from B
				movupd xmm5, [esi+16];		//Load from B
				movupd xmm6, [esi+32];		//Load from B
				movupd xmm7, [esi+48];		//Load from B

				psubw xmm0, xmm4;			//Multiply A*B
				psubw xmm1, xmm5;			//Multiply A*B
				psubw xmm2, xmm6;			//Multiply A*B
				psubw xmm3, xmm7;			//Multiply A*B

				movupd [ebx],	  xmm0;		//Move into C
				movupd [ebx+16],  xmm1;		//Move into C
				movupd [ebx+32],  xmm2;		//Move into C
				movupd [ebx+48],  xmm3;		//Move into C

				add edi, 64;
				add esi, 64;
				add ebx, 64;
				
			loop L1;							// Loop if not done

	ZERO:
			mov ecx, cnt2;
			jecxz ZERO1;

			L2:

				movupd xmm0, [edi];		//Load from A
				movupd xmm1, [esi];		//Load from B
				psubw  xmm0, xmm1;		//Multiply A*B
				movupd [ebx], xmm0;		//Move into C
				add edi, 16;
				add esi, 16;
				add ebx, 16;

			loop L2;	

	ZERO1:

			mov ecx, cnt3;
			jecxz ZERO2;

			mov eax, 0;

			L3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				sub ax, [esi];
				mov [ebx], ax;
				add esi, 2;
				add edi, 2;
				add ebx, 2;

			loop L3;

	ZERO2:

			EMMS;

		}//end __asm

	}
	else
	{

		cnt1 = cnt / 56;
		cnt2 = (cnt - (56*cnt1)) / 8;
		cnt3 = (cnt - (56*cnt1) - (8*cnt2));

		_asm 
		{
			//Set up for loop
			mov edi, A;	// Address of A, input1
			mov esi, B;	// Address of B, input2
			mov ebx, C; // Address of C, output1
			mov ecx, cnt1;	// Counter
			jecxz AZERO;

			AL1:

				movapd	xmm0, [edi];		//Load from A
				movapd	xmm1, [edi+16];		//Load from A
				movapd	xmm2, [edi+32];		//Load from A
				movapd	xmm3, [edi+48];		//Load from A
				movapd	xmm4, [edi+64];		//Load from A
				movapd	xmm5, [edi+80];		//Load from A
				movapd	xmm6, [edi+96];		//Load from A

				psubw xmm0, [esi];			//Load from B & multiply A*B
				psubw xmm1, [esi+16];		//Load from B & multiply A*B
				psubw xmm2, [esi+32];		//Load from B & multiply A*B
				psubw xmm3, [esi+48];		//Load from B & multiply A*B
				psubw xmm4, [esi+64];		//Load from B & multiply A*B
				psubw xmm5, [esi+80];		//Load from B & multiply A*B
				psubw xmm6, [esi+96];		//Load from B & multiply A*B

				movapd [ebx],	  xmm0;		//Move into A
				movapd [ebx+16],  xmm1;		//Move into A
				movapd [ebx+32],  xmm2;		//Move into A
				movapd [ebx+48],  xmm3;		//Move into A
				movapd [ebx+64],  xmm4;		//Move into A
				movapd [ebx+80],  xmm5;		//Move into A
				movapd [ebx+96],  xmm6;		//Move into A

				add edi, 112;
				add esi, 112;
				add ebx, 112;
				
			loop AL1;							// Loop if not done

	AZERO:
			mov ecx, cnt2;
			jecxz AZERO1;

			AL2:

				movapd  xmm0, [edi];		//Load from A
				psubw  xmm0, [esi];		//Load from B & add A+B
				movapd [ebx], xmm0;		//Move into A
				add edi, 16;
				add esi, 16;
				add ebx, 16;

			loop AL2;	

	AZERO1:

			mov ecx, cnt3;
			jecxz AZERO2;

			mov eax, 0;

			AL3:								//Really finish off loop with non SIMD instructions

				mov ax, [edi];
				sub ax, [esi];
				mov [ebx], ax;
				add esi, 2;
				add edi, 2;
				add ebx, 2;

			loop AL3;

	AZERO2:

			EMMS;

		}//end __asm

	}//end if
}

















void sse_conj(void *A, int cnt)
{

	int cnt1, cnt2;
	
	cnt1 = cnt/28;
	cnt2 = cnt - 28*cnt1;

	short M[8] = {1,-1,1,-1,1,-1,1,-1};

	if((int)A%16)
	{
		__asm 
		{
			
			//Set up for loop
			mov edi, A;			// Address of A	source1
			mov ecx, cnt1;		// Counter
			movq mm7, M;		// Move the multiply thingie
			movupd xmm7, M;		// Move the multiply thingie
			jecxz ZERO1;

			L1:

				movupd		xmm0, [edi];	//Load from A
				movupd		xmm1, [edi+16]; //Load from A
				movupd		xmm2, [edi+32]; //Load from A
				movupd		xmm3, [edi+48]; //Load from A
				movupd		xmm4, [edi+64]; //Load from A
				movupd		xmm5, [edi+80]; //Load from A
				movupd		xmm6, [edi+96]; //Load from A

				pmullw		xmm0, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm1, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm2, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm3, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm4, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm5, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm6, xmm7;		//Multiply to get [Re -Im Re -Im]

				movupd		[edi],    xmm0;	//Move into A
				movupd		[edi+16], xmm1;	//Move into A
				movupd		[edi+32], xmm2;	//Move into A
				movupd		[edi+48], xmm3;	//Move into A
				movupd		[edi+64], xmm4;	//Move into A
				movupd		[edi+80], xmm5;	//Move into A
				movupd		[edi+96], xmm6;	//Move into A

				add			edi, 112;		//Move in array
			
			loop L1;							// Loop if not done

	ZERO1:

			mov ecx, cnt2;		// Counter
			jecxz ZERO2;

			L2:

				movd	mm0, [edi];
				pmullw	mm0, mm7;
				movd	[edi], mm0;

				add		edi, 4;

			loop L2;

	ZERO2:

			EMMS;

		}//end __asm
	}
	else
	{
		__asm 
		{
			
			//Set up for loop
			mov edi, A;			// Address of A	source1
			mov ecx, cnt1;		// Counter
			movq mm7, M;		// Move the multiply thingie
			movupd xmm7, M;		// Move the multiply thingie
			jecxz AZERO1;

			AL1:

				movapd		xmm0, [edi];	//Load from A
				movapd		xmm1, [edi+16]; //Load from A
				movapd		xmm2, [edi+32]; //Load from A
				movapd		xmm3, [edi+48]; //Load from A
				movapd		xmm4, [edi+64]; //Load from A
				movapd		xmm5, [edi+80]; //Load from A
				movapd		xmm6, [edi+96]; //Load from A

				pmullw		xmm0, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm1, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm2, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm3, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm4, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm5, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm6, xmm7;		//Multiply to get [Re -Im Re -Im]

				movapd		[edi],    xmm0;	//Move into A
				movapd		[edi+16], xmm1;	//Move into A
				movapd		[edi+32], xmm2;	//Move into A
				movapd		[edi+48], xmm3;	//Move into A
				movapd		[edi+64], xmm4;	//Move into A
				movapd		[edi+80], xmm5;	//Move into A
				movapd		[edi+96], xmm6;	//Move into A

				add			edi, 112;		//Move in array
			
			loop AL1;							// Loop if not done

	AZERO1:

			mov ecx, cnt2;		// Counter
			jecxz AZERO2;

			AL2:

				movd	mm0, [edi];
				pmullw	mm0, mm7;
				movd	[edi], mm0;

				add		edi, 4;

			loop AL2;

	AZERO2:

			EMMS;

		}//end __asm

	}//end if

}



void sse_conj(void *A, void *C, int cnt)
{

	int cnt1, cnt2;
	
	cnt1 = cnt/28;
	cnt2 = cnt - 28*cnt1;

	short M[8] = {1,-1,1,-1,1,-1,1,-1};

	if(((int)A%16) || ((int)C%16))
	{
		__asm 
		{
			
			//Set up for loop
			mov edi, A;			// Address of A	source1
			mov esi, C;			// Address of C output1
			mov ecx, cnt1;		// Counter
			movq mm7, M;		// Move the multiply thingie
			movupd xmm7, M;		// Move the multiply thingie
			jecxz ZERO1;

			L1:

				movupd		xmm0, [edi];	//Load from A
				movupd		xmm1, [edi+16]; //Load from A
				movupd		xmm2, [edi+32]; //Load from A
				movupd		xmm3, [edi+48]; //Load from A
				movupd		xmm4, [edi+64]; //Load from A
				movupd		xmm5, [edi+80]; //Load from A
				movupd		xmm6, [edi+96]; //Load from A

				pmullw		xmm0, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm1, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm2, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm3, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm4, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm5, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm6, xmm7;		//Multiply to get [Re -Im Re -Im]

				movupd		[esi],    xmm0;	//Move into A
				movupd		[esi+16], xmm1;	//Move into A
				movupd		[esi+32], xmm2;	//Move into A
				movupd		[esi+48], xmm3;	//Move into A
				movupd		[esi+64], xmm4;	//Move into A
				movupd		[esi+80], xmm5;	//Move into A
				movupd		[esi+96], xmm6;	//Move into A

				add			edi, 112;		//Move in array
				add			esi, 112;
			
			loop L1;							// Loop if not done

	ZERO1:

			mov ecx, cnt2;		// Counter
			jecxz ZERO2;

			L2:

				movd	mm0, [edi];
				pmullw	mm0, mm7;
				movd	[esi], mm0;

				add		edi, 4;
				add		esi, 4;

			loop L2;

	ZERO2:

			EMMS;

		}//end __asm
	}
	else
	{
		__asm 
		{
			
			//Set up for loop
			mov edi, A;			// Address of A	source1
			mov esi, C;			// Address of C output1
			mov ecx, cnt1;		// Counter
			movq mm7, M;		// Move the multiply thingie
			movupd xmm7, M;		// Move the multiply thingie
			jecxz AZERO1;

			AL1:

				movapd		xmm0, [edi];	//Load from A
				movapd		xmm1, [edi+16]; //Load from A
				movapd		xmm2, [edi+32]; //Load from A
				movapd		xmm3, [edi+48]; //Load from A
				movapd		xmm4, [edi+64]; //Load from A
				movapd		xmm5, [edi+80]; //Load from A
				movapd		xmm6, [edi+96]; //Load from A

				pmullw		xmm0, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm1, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm2, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm3, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm4, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm5, xmm7;		//Multiply to get [Re -Im Re -Im]
				pmullw		xmm6, xmm7;		//Multiply to get [Re -Im Re -Im]

				movapd		[esi],    xmm0;	//Move into A
				movapd		[esi+16], xmm1;	//Move into A
				movapd		[esi+32], xmm2;	//Move into A
				movapd		[esi+48], xmm3;	//Move into A
				movapd		[esi+64], xmm4;	//Move into A
				movapd		[esi+80], xmm5;	//Move into A
				movapd		[esi+96], xmm6;	//Move into A

				add			edi, 112;		//Move in array
				add			esi, 112;		//Move in array
			
			loop AL1;							// Loop if not done

	AZERO1:

			mov ecx, cnt2;		// Counter
			jecxz AZERO2;

			AL2:

				movd	mm0, [edi];
				pmullw	mm0, mm7;
				movd	[esi], mm0;

				add		edi, 4;
				add		esi, 4;

			loop AL2;

	AZERO2:

			EMMS;

		}//end __asm

	}//end if

}



void sse_cmul(void *A, void *B, int cnt, int shift)
{

	int cnt1, cnt2;

	cnt1 = cnt / 4;
	cnt2 = cnt - (cnt1*4);

	short M[8] = {1,-1,1,1,1,-1,1,1};


	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov esi, B;			// Address of B	source2
		mov ecx, cnt1;		// Counter
		movupd xmm7, M;		// Move the multiply thingie
		movss  xmm6, shift;	// Move the round thingie
		jecxz ZERO1;

		L1:

			movlpd xmm0, [edi];		//Copy from A
			movlpd xmm1, [edi+8];	//Copy from A

			movlpd xmm3, [esi];		//Copy from B
			movlpd xmm4, [esi+8];	//Copy from B

			punpckldq xmm0, xmm0;	//Copy low 32 bits to high 32 bits
			punpckldq xmm1, xmm1;	//Copy low 32 bits to high 32 bits

			punpckldq xmm3, xmm3;	//Copy low 32 bits to high 32 bits
			punpckldq xmm4, xmm4;	//Copy low 32 bits to high 32 bits

			pshuflw xmm3, xmm3, 0x14; //Shuffle Low 64 bits to get [Re Im Im Re]
			pshuflw xmm4, xmm4, 0x14; //Shuffle Low 64 bits to get [Re Im Im Re]

			pshufhw xmm3, xmm3, 0x14; //Shuffle High 64 bits to get [Re Im Im Re]
			pshufhw xmm4, xmm4, 0x14; //Shuffle High 64 bits to get [Re Im Im Re]

			pmullw xmm3, xmm7;		//Multiply to get [Re Im -Im Re]
			pmullw xmm4, xmm7;		//Multiply to get [Re Im -Im Re]

			pmaddwd xmm0, xmm3;		//Complex multiply and add
			pmaddwd xmm1, xmm4;		//Complex multiply and add

			psrad xmm0, xmm6;		//Shift by X bits
			psrad xmm1, xmm6;		//Shift by X bits

			packssdw xmm0, xmm0;	//Get into low 64 bits
			packssdw xmm1, xmm1;	//Get into low 64 bits

			movsd [edi],   xmm0;	//Move into A
			movsd [edi+8], xmm1;	//Move into A

			add edi, 16;			//Move in array
			add esi, 16;			//Move in array
		
		loop L1;							// Loop if not done

ZERO1:

		mov ecx, cnt2;
		jecxz ZERO2;

L2:

			movlpd		xmm0, [edi];		//Copy from A
			movlpd		xmm1, [esi];		//Copy from B

			punpckldq	xmm0, xmm0;			//Copy low 32 bits to high 32 bits
			punpckldq	xmm1, xmm1;			//Copy low 32 bits to high 32 bits

			pshuflw		xmm1, xmm1, 0x14;	//Shuffle Low 64 bits to get [Re Im Im Re]
			pmullw		xmm1, xmm7;			//Multiply to get [Re Im -Im Re]
			pmaddwd		xmm0, xmm1;			//Complex multiply and add
			psrad		xmm0, xmm6;			//Shift by X bits
			packssdw	xmm0, xmm0;			//Get into low 32 bits
			movd		[edi], xmm0;		//Move into A

			add edi, 4;
			add esi, 4;

			loop L2;


ZERO2:
		EMMS;

	}


}



void sse_cmul(void *A, void *B, void *C, int cnt, int shift)
{

	int cnt1, cnt2;

	cnt1 = cnt / 4;
	cnt2 = cnt - (cnt1*4);

	short M[8] = {1,-1,1,1,1,-1,1,1};


	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov esi, B;			// Address of B	source2
		mov ebx, C;			// Address of C output1
		mov ecx, cnt1;		// Counter
		movupd xmm7, M;		// Move the multiply thingie
		movss  xmm6, shift;	// Move the round thingie
		jecxz ZERO1;

		L1:

			movlpd xmm0, [edi];		//Copy from A
			movlpd xmm1, [edi+8];	//Copy from A

			movlpd xmm3, [esi];		//Copy from B
			movlpd xmm4, [esi+8];	//Copy from B

			punpckldq xmm0, xmm0;	//Copy low 32 bits to high 32 bits
			punpckldq xmm1, xmm1;	//Copy low 32 bits to high 32 bits

			punpckldq xmm3, xmm3;	//Copy low 32 bits to high 32 bits
			punpckldq xmm4, xmm4;	//Copy low 32 bits to high 32 bits

			pshuflw xmm3, xmm3, 0x14; //Shuffle Low 64 bits to get [Re Im Im Re]
			pshuflw xmm4, xmm4, 0x14; //Shuffle Low 64 bits to get [Re Im Im Re]

			pshufhw xmm3, xmm3, 0x14; //Shuffle High 64 bits to get [Re Im Im Re]
			pshufhw xmm4, xmm4, 0x14; //Shuffle High 64 bits to get [Re Im Im Re]

			pmullw xmm3, xmm7;		//Multiply to get [Re Im -Im Re]
			pmullw xmm4, xmm7;		//Multiply to get [Re Im -Im Re]

			pmaddwd xmm0, xmm3;		//Complex multiply and add
			pmaddwd xmm1, xmm4;		//Complex multiply and add

			psrad xmm0, xmm6;		//Shift by X bits
			psrad xmm1, xmm6;		//Shift by X bits

			packssdw xmm0, xmm0;	//Get into low 64 bits
			packssdw xmm1, xmm1;	//Get into low 64 bits

			movsd [ebx],   xmm0;	//Move into C
			movsd [ebx+8], xmm1;	//Move into C

			add edi, 16;			//Move in array
			add esi, 16;			//Move in array
			add ebx, 16;			//Move in array
		
		loop L1;							// Loop if not done

ZERO1:

		mov ecx, cnt2;
		jecxz ZERO2;

L2:

			movlpd		xmm0, [edi];		//Copy from A
			movlpd		xmm1, [esi];		//Copy from B

			punpckldq	xmm0, xmm0;			//Copy low 32 bits to high 32 bits
			punpckldq	xmm1, xmm1;			//Copy low 32 bits to high 32 bits

			pshuflw		xmm1, xmm1, 0x14;	//Shuffle Low 64 bits to get [Re Im Im Re]
			pmullw		xmm1, xmm7;			//Multiply to get [Re Im -Im Re]
			pmaddwd		xmm0, xmm1;			//Complex multiply and add
			psrad		xmm0, xmm6;			//Shift by X bits
			packssdw	xmm0, xmm0;			//Get into low 32 bits
			movd		[ebx], xmm0;		//Move into A

			add edi, 4;
			add esi, 4;
			add ebx, 4;

			loop L2;


ZERO2:
		EMMS;

	}


}


void sse_crot(void *A, void *B, int cnt, int shift)
{

	int cnt1;
	int cnt2;

	cnt1 = cnt/8;
	cnt2 = cnt - (cnt1*8);

	short M[8] = {1,-1,1,1,1,-1,1,1};			//To get mmx register into Re Im -Im Re format

	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov esi, B;			// Address of B	source2
		mov ecx, cnt1;		// Counter
		movupd xmm7, M;		// Move the multiply thingie
		movss  xmm6, shift;	// Move the round thingie

		//Set up the rotation register
		movss	xmm1, [esi];
		pshuflw	xmm1, xmm1, 0x14;	//Shuffle to get [Re Im Im Re]
		pmullw	xmm1, xmm7;			//Multiply to get [Re -Im Im Re]
		pshufd	xmm7, xmm1, 0x44;	//Now in mm7

		jecxz ZERO1;

		L1:

			movlpd xmm0, [edi];		//Copy from A
			movlpd xmm1, [edi+8];	//Copy from A
			movlpd xmm2, [edi+16];	//Copy from A
			movlpd xmm3, [edi+24];	//Copy from A

			punpckldq	xmm0, xmm0;	//Copy low 32 bits to high 32 bits
			punpckldq	xmm1, xmm1;	//Copy low 32 bits to high 32 bits
			punpckldq	xmm2, xmm2;	//Copy low 32 bits to high 32 bits
			punpckldq	xmm3, xmm3;	//Copy low 32 bits to high 32 bits

			pmaddwd		xmm0, xmm7;	//Complex multiply and add
			pmaddwd		xmm1, xmm7;	//Complex multiply and add
			pmaddwd		xmm2, xmm7;	//Complex multiply and add
			pmaddwd		xmm3, xmm7;	//Complex multiply and add

			psrad		xmm0, xmm6;	//Shift by X bits
			psrad		xmm1, xmm6;	//Shift by X bits
			psrad		xmm2, xmm6;	//Shift by X bits
			psrad		xmm3, xmm6;	//Shift by X bits

			packssdw	xmm0, xmm0;	//Get into low 32 bits
			packssdw	xmm1, xmm1;	//Get into low 32 bits
			packssdw	xmm2, xmm2;	//Get into low 32 bits
			packssdw	xmm3, xmm3;	//Get into low 32 bits

			movsd [edi],    xmm0;	//Move into A
			movsd [edi+8],  xmm1;	//Move into A
			movsd [edi+16], xmm2;	//Move into A
			movsd [edi+24], xmm3;	//Move into A

			add			edi, 32;	//Move in array
		
		loop L1;							// Loop if not done

ZERO1:

		mov ecx, cnt2;
		jecxz ZERO2;

		L2:

			movss		xmm0, [edi];	//Copy from A

			punpckldq	xmm0, xmm0;		//Copy low 32 bits to high 32 bits
			pmaddwd		xmm0, xmm7;		//Complex multiply and add
			psrad		xmm0, xmm6;		//Shift by X bits
			packssdw	xmm0, xmm0;		//Get into low 32 bits

			movss		[edi], xmm0;	//Move into A

			add			edi, 4;			//Move in array
		
		loop L2;						// Loop if not done

ZERO2:

		EMMS;

	}


}



void sse_crot(void *A, void *B, void *C, int cnt, int shift)
{

	int cnt1;
	int cnt2;

	cnt1 = cnt/8;
	cnt2 = cnt - (cnt1*8);

	short M[8] = {1,-1,1,1,1,-1,1,1};			//To get mmx register into Re Im -Im Re format

	__asm 
	{
		
		//Set up for loop
		mov edi, A;			// Address of A	source1
		mov esi, B;			// Address of B	source2
		mov ebx, C;			// Address of C input1
		mov ecx, cnt1;		// Counter
		movupd xmm7, M;		// Move the multiply thingie
		movss  xmm6, shift;	// Move the round thingie

		//Set up the rotation register
		movss	xmm1, [esi];
		pshuflw	xmm1, xmm1, 0x14;	//Shuffle to get [Re Im Im Re]
		pmullw	xmm1, xmm7;			//Multiply to get [Re -Im Im Re]
		pshufd	xmm7, xmm1, 0x44;	//Now in mm7

		jecxz ZERO1;

		L1:

			movlpd xmm0, [edi];		//Copy from A
			movlpd xmm1, [edi+8];	//Copy from A
			movlpd xmm2, [edi+16];	//Copy from A
			movlpd xmm3, [edi+24];	//Copy from A

			punpckldq	xmm0, xmm0;	//Copy low 32 bits to high 32 bits
			punpckldq	xmm1, xmm1;	//Copy low 32 bits to high 32 bits
			punpckldq	xmm2, xmm2;	//Copy low 32 bits to high 32 bits
			punpckldq	xmm3, xmm3;	//Copy low 32 bits to high 32 bits

			pmaddwd		xmm0, xmm7;	//Complex multiply and add
			pmaddwd		xmm1, xmm7;	//Complex multiply and add
			pmaddwd		xmm2, xmm7;	//Complex multiply and add
			pmaddwd		xmm3, xmm7;	//Complex multiply and add

			psrad		xmm0, xmm6;	//Shift by X bits
			psrad		xmm1, xmm6;	//Shift by X bits
			psrad		xmm2, xmm6;	//Shift by X bits
			psrad		xmm3, xmm6;	//Shift by X bits

			packssdw	xmm0, xmm0;	//Get into low 32 bits
			packssdw	xmm1, xmm1;	//Get into low 32 bits
			packssdw	xmm2, xmm2;	//Get into low 32 bits
			packssdw	xmm3, xmm3;	//Get into low 32 bits

			movsd [ebx],    xmm0;	//Move into A
			movsd [ebx+8],  xmm1;	//Move into A
			movsd [ebx+16], xmm2;	//Move into A
			movsd [ebx+24], xmm3;	//Move into A

			add			edi, 32;	//Move in array
			add			ebx, 32;
		
		loop L1;							// Loop if not done

ZERO1:

		mov ecx, cnt2;
		jecxz ZERO2;

		L2:

			movss		xmm0, [edi];		//Copy from A

			punpckldq	xmm0, xmm0;	//Copy low 32 bits to high 32 bits
			pmaddwd		xmm0, xmm7;	//Complex multiply and add
			psrad		xmm0, xmm6;	//Shift by X bits
			packssdw	xmm0, xmm0;	//Get into low 32 bits

			movss		[ebx], xmm0;	//Move into A

			add			edi, 4;		//Move in array
			add			ebx, 4;
		
		loop L2;					// Loop if not done

ZERO2:

		EMMS;

	}


}



void sse_qnt(void *A, int cnt)
{

	__int64 hash1 = 0x8000800080008000; //to save the sign bit
	__int64 hash2 = 0x0001000100010001; //add the "1" magnitude
	unsigned short	shash1 = 0x8000;
	unsigned short	shash2 = 0xfffe;
	int cnt1;
	int cnt2;

	cnt1 = cnt / 40;						
	cnt2 = (cnt - (40*cnt1));				

	if(((int)A%16))
	{
		__asm 
		{
			
			// Set up for loop
			mov edi, A;			// Address of A
			mov ecx, cnt1;		// Counter
			movq xmm6, [hash1];	//move in the sign bit hash
			movq xmm7, [hash2];	//move in the magnitude hash
			pshufd	xmm6, xmm6, 0x44;	
			pshufd	xmm7, xmm7, 0x44;	

			jecxz ZERO;

			L1:

				movupd xmm0, [edi];
				movupd xmm1, [edi+16];
				movupd xmm2, [edi+32];
				movupd xmm3, [edi+48];
				movupd xmm4, [edi+64];

				pand xmm0, xmm6;		//Clear everything except the sign bit
				pand xmm1, xmm6;		//Clear everything except the sign bit
				pand xmm2, xmm6;		//Clear everything except the sign bit
				pand xmm3, xmm6;		//Clear everything except the sign bit
				pand xmm4, xmm6;		//Clear everything except the sign bit

				pcmpeqw xmm0, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm1, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm2, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm3, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm4, xmm6;		//Is the sign bit 1 or 0

				por		xmm0, xmm7;		//Add the magnitude
				por		xmm1, xmm7;		//Add the magnitude
				por		xmm2, xmm7;		//Add the magnitude
				por		xmm3, xmm7;		//Add the magnitude
				por		xmm4, xmm7;		//Add the magnitude

				movupd	[edi],	  xmm0;	//Move back to the array
				movupd	[edi+16], xmm1;	//Move back to the array
				movupd	[edi+32], xmm2;	//Move back to the array
				movupd	[edi+48], xmm3;	//Move back to the array
				movupd	[edi+64], xmm4;	//Move back to the array

				add edi, 80;

			loop L1;					// Loop if not done

	ZERO:
				mov ecx, cnt2;			// Finish off operation with second loop
				jecxz ZERO1;

				mov ax, shash1;
				mov bx, shash2;

			L2:

				mov dx, [edi];
				and dx, ax;
				cmp dx, ax;
				cmove dx, bx;
				add dx, 0x0001;
				mov [edi], dx;

				add edi, 2;

			loop L2;

				
	ZERO1:	

			EMMS;							// done with MMX

		
		}
	}
	else
	{
		__asm 
		{
			
			// Set up for loop
			mov edi, A;			// Address of A
			mov ecx, cnt1;		// Counter
			movq xmm6, [hash1];	//move in the sign bit hash
			movq xmm7, [hash2];	//move in the magnitude hash
			pshufd	xmm6, xmm6, 0x44;	
			pshufd	xmm7, xmm7, 0x44;	

			jecxz AZERO;

			AL1:

				movapd xmm0, [edi];
				movapd xmm1, [edi+16];
				movapd xmm2, [edi+32];
				movapd xmm3, [edi+48];
				movapd xmm4, [edi+64];

				pand xmm0, xmm6;		//Clear everything except the sign bit
				pand xmm1, xmm6;		//Clear everything except the sign bit
				pand xmm2, xmm6;		//Clear everything except the sign bit
				pand xmm3, xmm6;		//Clear everything except the sign bit
				pand xmm4, xmm6;		//Clear everything except the sign bit

				pcmpeqw xmm0, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm1, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm2, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm3, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm4, xmm6;		//Is the sign bit 1 or 0

				por		xmm0, xmm7;		//Add the magnitude
				por		xmm1, xmm7;		//Add the magnitude
				por		xmm2, xmm7;		//Add the magnitude
				por		xmm3, xmm7;		//Add the magnitude
				por		xmm4, xmm7;		//Add the magnitude

				movapd	[edi],	  xmm0;	//Move back to the array
				movapd	[edi+16], xmm1;	//Move back to the array
				movapd	[edi+32], xmm2;	//Move back to the array
				movapd	[edi+48], xmm3;	//Move back to the array
				movapd	[edi+64], xmm4;	//Move back to the array

				add edi, 80;

			loop AL1;					// Loop if not done

	AZERO:
				mov ecx, cnt2;			// Finish off operation with second loop
				jecxz AZERO1;

				mov ax, shash1;
				mov bx, shash2;

			AL2:

				mov dx, [edi];
				and dx, ax;
				cmp dx, ax;
				cmove dx, bx;
				add dx, 0x0001;
				mov [edi], dx;

				add edi, 2;

			loop AL2;

				
	AZERO1:	

			EMMS;							// done with MMX

		
		}
	}




}


void sse_qnt(void *A, void *C, int cnt)
{

	__int64 hash1 = 0x8000800080008000; //to save the sign bit
	__int64 hash2 = 0x0001000100010001; //add the "1" magnitude
	unsigned short	shash1 = 0x8000;
	unsigned short	shash2 = 0xfffe;
	int cnt1;
	int cnt2;

	cnt1 = cnt / 40;						
	cnt2 = (cnt - (40*cnt1));				

	if(((int)A%16) || ((int)C%16))
	{
		__asm 
		{
			
			// Set up for loop
			mov edi, A;			// Address of A
			mov esi, C;			// Address of C
			mov ecx, cnt1;		// Counter
			movq xmm6, [hash1];	//move in the sign bit hash
			movq xmm7, [hash2];	//move in the magnitude hash
			pshufd	xmm6, xmm6, 0x44;	
			pshufd	xmm7, xmm7, 0x44;	
			jecxz ZERO;


			L1:

				movupd xmm0, [edi];
				movupd xmm1, [edi+16];
				movupd xmm2, [edi+32];
				movupd xmm3, [edi+48];
				movupd xmm4, [edi+64];

				pand xmm0, xmm6;		//Clear everything except the sign bit
				pand xmm1, xmm6;		//Clear everything except the sign bit
				pand xmm2, xmm6;		//Clear everything except the sign bit
				pand xmm3, xmm6;		//Clear everything except the sign bit
				pand xmm4, xmm6;		//Clear everything except the sign bit

				pcmpeqw xmm0, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm1, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm2, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm3, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm4, xmm6;		//Is the sign bit 1 or 0

				por		xmm0, xmm7;		//Add the magnitude
				por		xmm1, xmm7;		//Add the magnitude
				por		xmm2, xmm7;		//Add the magnitude
				por		xmm3, xmm7;		//Add the magnitude
				por		xmm4, xmm7;		//Add the magnitude

				movupd	[esi],	  xmm0;	//Move back to the array
				movupd	[esi+16], xmm1;	//Move back to the array
				movupd	[esi+32], xmm2;	//Move back to the array
				movupd	[esi+48], xmm3;	//Move back to the array
				movupd	[esi+64], xmm4;	//Move back to the array

				add edi, 80;
				add esi, 80;

			loop L1;					// Loop if not done

	ZERO:
				mov ecx, cnt2;			// Finish off operation with second loop
				jecxz ZERO1;

				mov ax, shash1;
				mov bx, shash2;

			L2:

				mov dx, [edi];
				and dx, ax;
				cmp dx, ax;
				cmove dx, bx;
				add dx, 0x0001;
				mov [esi], dx;

				add edi, 2;
				add esi, 2;

			loop L2;

				
	ZERO1:	

			EMMS;							// done with MMX

		
		}
	}
	else
	{
		__asm 
		{
			
			// Set up for loop
			mov edi, A;			// Address of A
			mov esi, C;			// Address of C
			mov ecx, cnt1;		// Counter
			movq xmm6, [hash1];	//move in the sign bit hash
			movq xmm7, [hash2];	//move in the magnitude hash
			pshufd	xmm6, xmm6, 0x44;	
			pshufd	xmm7, xmm7, 0x44;	
			jecxz AZERO;


			AL1:

				movapd xmm0, [edi];
				movapd xmm1, [edi+16];
				movapd xmm2, [edi+32];
				movapd xmm3, [edi+48];
				movapd xmm4, [edi+64];

				pand xmm0, xmm6;		//Clear everything except the sign bit
				pand xmm1, xmm6;		//Clear everything except the sign bit
				pand xmm2, xmm6;		//Clear everything except the sign bit
				pand xmm3, xmm6;		//Clear everything except the sign bit
				pand xmm4, xmm6;		//Clear everything except the sign bit

				pcmpeqw xmm0, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm1, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm2, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm3, xmm6;		//Is the sign bit 1 or 0
				pcmpeqw xmm4, xmm6;		//Is the sign bit 1 or 0

				por		xmm0, xmm7;		//Add the magnitude
				por		xmm1, xmm7;		//Add the magnitude
				por		xmm2, xmm7;		//Add the magnitude
				por		xmm3, xmm7;		//Add the magnitude
				por		xmm4, xmm7;		//Add the magnitude

				movapd	[esi],	  xmm0;	//Move back to the array
				movapd	[esi+16], xmm1;	//Move back to the array
				movapd	[esi+32], xmm2;	//Move back to the array
				movapd	[esi+48], xmm3;	//Move back to the array
				movapd	[esi+64], xmm4;	//Move back to the array

				add edi, 80;
				add esi, 80;

			loop AL1;					// Loop if not done

	AZERO:
				mov ecx, cnt2;			// Finish off operation with second loop
				jecxz AZERO1;

				mov ax, shash1;
				mov bx, shash2;

			AL2:

				mov dx, [edi];
				and dx, ax;
				cmp dx, ax;
				cmove dx, bx;
				add dx, 0x0001;
				mov [esi], dx;

				add edi, 2;
				add esi, 2;

			loop AL2;

				
	AZERO1:	

			EMMS;							// done with MMX

		
		}
	}




}


